# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A class to handle continuum absorption (CIA)
"""

import os.path
import h5py
import numpy as np
from .settings import Settings
from .util.interp import linear_interpolation, interp_ind_weights
from .util.spectral_object import Spectral_object
from .util.cst import PI


class Atable(Spectral_object):
    """A class to handle aerosol optical properties in table form.
    """

    def __init__(self, *filename_filters, filename=None, aerosol_name=None, search_path=None,
            mks=False, remove_zeros=False):
        """Initialization for Cia_tables.

        Parameters
        ----------
            filename: str, optional
                Relative or absolute name of the file to be loaded. 
            filename_filters: sequence of string
                As many strings as necessary to uniquely define
                a file in the global search path defined in
                :class:`~exo_k.settings.Settings`.
                This path will be searched for a file
                with all the filename_filters in the name.
                The filename_filters can contain '*'.

        Other Parameters
        ----------------
            search_path: str, optional
                If search_path is provided,
                it locally overrides the global _search_path
                in :class:`~exo_k.settings.Settings`
                and only files in search_path are returned.
        """
        self._init_empty()
        self._settings=Settings()
        if filename is not None:
            self.filename=filename
        elif filename_filters :
            # a none empty sequence returns a True in a conditional statement
            self.filename=self._settings.list_files(*filename_filters,
                only_one=True, search_path=search_path)[0]

        if self.filename is not None:
            if self.filename.lower().endswith(('h5','hdf5')):
                self.read_hdf5(self.filename, aerosol_name=aerosol_name)
            elif self.filename.lower().endswith('.dat'):
                self.read_LMDZ(self.filename, aerosol_name=aerosol_name)
            else:
                raise RuntimeError('Aerosol optical property file extension not known.')
        if self.ext_coeff is not None:
            if self._settings._convert_to_mks or mks: self.convert_to_mks()
            if remove_zeros: self.remove_zeros()

    def _init_empty(self):
        """Initializes attributes to none.
        """
        self.aerosol_name=None
        self.wns=None
        self.wnedges=None
        self.wn_unit='cm^-1'
        self.ext_coeff=None
        self.single_scat_alb=None
        self.asymmetry_factor=None
        self.r_eff_grid=None
        self.r_eff_unit=None
        self.Nw=None
        self.Nr=None
        self.filename=None


    def read_LMDZ(self, filename, aerosol_name=None):
        """Reads LMDZ like optical propertt files.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
        """
        with open(filename, 'r') as file:
            _=file.readline()
            self.Nw=int(file.readline())
            _=file.readline()
            self.Nr=int(file.readline())
            _=file.readline()
            self.wns=self._read_array(file, self.Nw)
            self.wns=.01/self.wns # conversion from m to cm^-1
            _=file.readline()
            self.r_eff_grid=self._read_array(file, self.Nr, revert=False)
            # read ext coeff
            _=file.readline()
            self.ext_coeff=self._read_arrays(file, self.Nw, self.Nr)
            # read albedo
            _=file.readline()
            self.single_scat_alb=self._read_arrays(file, self.Nw, self.Nr)
            # read asymmetry factor
            _=file.readline()
            self.asymmetry_factor=self._read_arrays(file, self.Nw, self.Nr)
        self.r_eff_unit='m'
        if aerosol_name:
            self.aerosol_name=aerosol_name
        else:
            self.aerosol_name=os.path.basename(filename).split('.')[0]




    def _read_array(self, file, Nvalue, Nline=None, revert=True):
        """Reads an array in a optical property LMDZ file. 
        Assumes that the arrays are arranged 5 values per line. 

        Parameters
        ----------
            file: file stream
                File to be read.
            Nvalue: int
                Number of values to be read. 
            revert: boolean (optional)
                Whether or not to revert the array (because most arrays are
                ordered by decreasing wavenumber).

        Returns
        -------
            Array
                A numpy array with the values.
        """
        
        if Nline is None:
            Nline=Nvalue//5
            if Nvalue%5 != 0:
                Nline+=1
        new_array=[]
        for _ in range(Nline):
            line=[float(i) for i in file.readline().split()]
            new_array+=line
        if revert :
            new_array=np.array(new_array)[::-1]
        else:
            new_array=np.array(new_array)
        return new_array

    def _read_arrays(self, file, Nvalue, Narray):
        """Reads an array in a optical property LMDZ file. 
        Assumes that the arrays are arranged 5 values per line. 

        Parameters
        ----------
            file: file stream
                File to be read.
            Nvalue: int
                Number of values to be read in each array. 
            Narray: int
                Number of arrays to be read. 

        Returns
        -------
            Array
                A numpy array with the values.
        """
        
        Nline=Nvalue//5
        if Nvalue%5 != 0:
            Nline+=1
        new_array=np.zeros((Narray,Nvalue))
        for ii in range(Narray):
            file.readline()
            new_array[ii]=self._read_array(file, Nvalue, Nline=Nline)
        return new_array

    def read_hdf5(self, filename, aerosol_name=None):
        """Reads hdf5 cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
        """
        f = h5py.File(filename, 'r')
        if aerosol_name:
            self.aerosol_name=aerosol_name
        else:
            self.aerosol_name=f['aerosol_name'][()]
        self.wns=f['bin_centers'][...]
        if 'units' in f['bin_centers'].attrs:
            self.wn_unit=f['bin_centers'].attrs['units']
        self.ext_coeff=f['ext_coeff'][...]
        self.single_scat_alb=f['single_scat_alb'][...]
        self.asymmetry_factor=f['asymmetry_factor'][...]
        self.r_eff_grid=f['r_eff'][...]
        self.r_eff_unit=f['r_eff'].attrs['units']
        f.close()  
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nr=self.r_eff_grid.size
        self.Nw=self.wns.size
          
    def write_hdf5(self, filename):
        """Writes hdf5 cia files.

        Parameters
        ----------
            filename: str
                Name of the file to be written.
        """
        if not filename.lower().endswith(('.hdf5', '.h5')):
            filename=filename+'.hdf5'
        f = h5py.File(filename, 'w')
        compression="gzip"
        f.create_dataset("aerosol_name", data=self.aerosol_name)
        f.create_dataset("bin_centers", data=self.wns, compression=compression)
        f.create_dataset("r_eff", data=self.r_eff_grid, compression=compression)
        f.create_dataset("ext_coeff", data=self.ext_coeff, compression=compression)
        f.create_dataset("single_scat_alb", data=self.single_scat_alb, compression=compression)
        f.create_dataset("asymmetry_factor", data=self.asymmetry_factor, compression=compression)
        f["bin_centers"].attrs["units"] = self.wn_unit
        f["r_eff"].attrs["units"] = self.r_eff_unit
        f.close()    

    def sample(self, wngrid, remove_zeros=False, use_grid_filter=False,
            sample_all_vars=False, **kwargs):
        """Method to re sample a Atable to a new grid of wavenumbers (in place)

        Parameters
        ----------
            wngrid : array
                new wavenumber grid (cm-1)
            use_grid_filter: boolean, optional
                If true, the table is sampled only within the boundaries
                of its current wavenumber grid. The coefficients are set to zero elswere
                (except if remove_zeros is set to True).
                If false, the values at the boundaries are used when sampling outside the grid.
            sample_all_vars: boolean, optional
                Whether to sample the single_scattering albedo and asymmetry_factor as well. 
        """
        wngrid=np.array(wngrid)
        Nnew=wngrid.size
        if use_grid_filter:
            wngrid_filter = np.where((wngrid <= self.wns[-1]) & (wngrid >= self.wns[0]))[0]
        else:
            wngrid_filter = np.ones(Nnew,dtype=bool)
        new_values=np.zeros((self.Nr, Nnew))
        for iR in range(self.Nr):
            tmp=self.ext_coeff[iR,:]
            new_values[iR,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
        self.ext_coeff=new_values
        if sample_all_vars :
            new_values=np.zeros((self.Nr, Nnew))
            new_values2=np.zeros((self.Nr, Nnew))
            for iR in range(self.Nr):
                tmp=self.single_scat_alb[iR,:]
                new_values[iR,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
                tmp=self.asymmetry_factor[iR,:]
                new_values2[iR,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
            self.single_scat_alb=new_values
            self.asymmetry_factor=new_values2
        self.wns=wngrid
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nw=Nnew
        if remove_zeros : self.remove_zeros(**kwargs)

    def sample_cp(self, wngrid, **kwargs):
        """Creates a copy of the object before resampling it.

        Parameters
        ----------
            See sample method for details. 

        Returns
        -------
            :class:`Atable` object
                the re-sampled :class:`Atable`
        """
        res=self.copy()
        res.sample(wngrid, **kwargs)
        return res

    def interpolate_optical_properties(self, r_array=None, var_type=0, 
            log_interp=None, wngrid_limit=None):
        """interpolate_cia interpolates the kdata at on a given temperature profile. 

        Parameters
        ----------
            r_array: float or array
                Effective radius array to interpolate to.
                If a float is given, it is interpreted as an array of size 1.
            data_id: int
                type of data to interpolate:
                  * 0 is extinction coefficient
                  * 1 is single scattering albedo
                  * 2 is asymmetry factor
            wngrid_limit: array, optional
                If an array is given, interpolates only within this array.
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).
        """
        if hasattr(r_array, "__len__"):
            r_array=np.array(r_array)
        else:
            r_array=np.array([r_array])
        rind,rweight=interp_ind_weights(r_array,self.r_eff_grid)
        if wngrid_limit is None:
            wngrid_filter = slice(None)
            Nw=self.Nw
        else:
            wngrid_limit=np.array(wngrid_limit)
            wngrid_filter = np.where((self.wnedges > wngrid_limit.min()) & (
                self.wnedges <= wngrid_limit.max()))[0][:-1]
            Nw=wngrid_filter.size
        res=np.zeros((rind.size,Nw))
        if var_type == 0:
            data_to_interp=self.ext_coeff
        elif var_type ==1:
            data_to_interp=self.single_scat_alb
        else:
            data_to_interp=self.asymmetry_factor
        if log_interp is None: log_interp=self._settings._log_interp
        if log_interp:
            for ii in range(rind.size):
                kc_t1=np.log(data_to_interp[rind[ii]][wngrid_filter].ravel())
                kc_t0=np.log(data_to_interp[rind[ii]-1][wngrid_filter].ravel())
                res[ii]=linear_interpolation(kc_t0, kc_t1, rweight[ii])
            return np.exp(res)
        else:
            for ii in range(rind.size):
                kc_t1=data_to_interp[rind[ii]][wngrid_filter].ravel()
                kc_t0=data_to_interp[rind[ii]-1][wngrid_filter].ravel()
                res[ii]=linear_interpolation(kc_t0, kc_t1, rweight[ii])
            return res

    def cross_section(self, r_array, wngrid_limit=None, log_interp=None):
        """Computes the cross section due to the aerosol in area per particles.

        Parameters
        ----------
            r_array: float or array
                Effective radius array to interpolate to.
                If a float is given, it is interpreted as an array of size 1.

        Other Parameters
        ----------------
            wngrid_limit: array, optional
                If an array is given, interpolates only within this array.
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).
        """
        if hasattr(r_array, "__len__"):
            r_array=np.array(r_array)
        else:
            r_array=np.array([r_array])
        area=PI*r_array**2
        return (area*self.interpolate_optical_properties(r_array=r_array,
            wngrid_limit=wngrid_limit, log_interp=log_interp).transpose()).transpose()

    def absorption_coefficient(self, r_array, n_density, wngrid_limit=None, log_interp=None):
        """Computes the cross section due to the aerosol in area per particles.

        Parameters
        ----------
            r_array: float or 1d array
                Effective radius array to interpolate to.
                If a float is given, it is interpreted as an array of size 1.
            n_density: float or 1d array (same dim as r_array)
                Number density of aerosol

        Other Parameters
        ----------------
            wngrid_limit: array, optional
                If an array is given, interpolates only within this array.
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).
        """
        if hasattr(r_array, "__len__"):
            r_array=np.array(r_array)
        else:
            r_array=np.array([r_array])
        factor=np.array(n_density)*PI*r_array**2
        return (factor*self.interpolate_optical_properties(r_array=r_array,
            wngrid_limit=wngrid_limit, log_interp=log_interp).transpose()).transpose()

    def plot_spectrum(self, ax, r=1.e-6, x_axis='wls', xscale=None, yscale=None, 
            var_type=0, **kwarg):
        """Plot the spectrum for a given point

        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            r: float
                Effective radius (m)
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        toplot=self.interpolate_optical_properties(r_array=r, var_type=var_type)[0]
        if var_type==0:
            ax.set_ylabel('Ext. coeff')
        elif var_type==1:
            ax.set_ylabel('Single Scat. Albedo')
        else:
            ax.set_ylabel('Ass. factor')
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        ax.grid(True)
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    def convert_to_mks(self):
        """Converts units to MKS
        """
        pass

    def rindex(self, r):
        """Finds the index corresponding to the given radius r
        (units must be the same as the ktable)
        """
        return min(np.searchsorted(self.r_eff_grid,r),self.Nr-1)

    def remove_zeros(self, deltalog_min_value=0.):
        """Finds zeros in the ext_coeff and set them to (10.^-deltalog_min_value)
        times the minimum positive value in the table.
        This is to be able to work in logspace. 
        """
        mask = np.zeros(self.ext_coeff.shape,dtype=bool)
        mask[np.nonzero(self.ext_coeff)] = True
        minvalue=np.amin(self.ext_coeff[mask])
        self.ext_coeff[~mask]=minvalue/(10.**deltalog_min_value)
        
    def __repr__(self):
        """Method to output header
        """
        output="""
        file          : {file}
        aerosol name  : {name}
        reff grid (mu): {r}
        wavenumber    : {w}
        ext coeff     : {ext_coeff}
        albedo        : {alb}
        asymmetry     : {ass}
        """.format(file=self.filename,name=self.aerosol_name, r=self.r_eff_grid,\
            w=self.wns,ext_coeff=self.ext_coeff, \
            alb=self.single_scat_alb, ass=self.asymmetry_factor)
        return output

    def __getitem__(self, key):
        """Overrides getitem.
        """
        return self.ext_coeff[key]

    def copy(self):
        """Creates a new instance of :class:`Atable` object and (deep) copies data into it
        """
        res=Atable()
        res.aerosol_name = self.aerosol_name
        res.wns = np.copy(self.wns)
        res.wn_unit = self.wn_unit
        res.wnedges = np.copy(self.wnedges)
        res.r_eff_grid = np.copy(self.r_eff_grid)
        res.r_eff_unit = self.r_eff_unit
        res.ext_coeff = np.copy(self.ext_coeff)
        res.single_scat_alb = np.copy(self.single_scat_alb)
        res.asymmetry_factor = np.copy(self.asymmetry_factor)
        res.Nr = self.Nr
        res.Nw = self.Nw
        res.filename = self.filename
        return res

def combine_tables(aerosol_name, *atables):
    """Combine several tables representing the same aerosol type for different wavelength range.
    """
    r_eff_grid=atables[0].r_eff_grid
    order=np.argsort([atable.wns[0] for atable in atables])
    ordered_atables=np.array(atables)[order]
    #print(ordered_atables)
    res=ordered_atables[0].copy()
    r_eff_grid=res.r_eff_grid
    wns_max=res.wns[-1]
    for atable in ordered_atables[1:]:
        if np.any(atable.r_eff_grid != r_eff_grid):
            raise RuntimeError('Not the same r_eff grid for all tables')
        print(atable.wns[0], wns_max)
        if atable.wns[0]<wns_max:
            raise RuntimeError('Overlapping wavenumber range')
        else:
            wns_max=atable.wns[-1]
    res.wns=np.concatenate([atable.wns[:-1] for atable in ordered_atables])
    res.Nw=res.wns.size
    res.ext_coeff=np.concatenate([atable.ext_coeff[:,:-1] for atable in ordered_atables], axis=1)
    res.single_scat_alb=np.concatenate(
        [atable.single_scat_alb[:,:-1] for atable in ordered_atables], axis=1)
    res.asymmetry_factor=np.concatenate(
        [atable.asymmetry_factor[:,:-1] for atable in ordered_atables], axis=1)
    res.aerosol_name=aerosol_name
    return res



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Class for Rayleigh opacties. 
"""
import numpy as np
from .util.cst import PI, KBOLTZ
from .util.singleton import Singleton

class Rayleigh(Singleton):
    """Class to compute Rayleigh opacities
    """

    def init(self, *args, **kwds):
        """Initializes various parameters for Rayleigh computations
        """
        N_std=1.e5/(KBOLTZ*273.15)
        self._mult_factor=24.*PI**3/(N_std)**2
        self._mult_factor=self._mult_factor*100.**4
        # last 100.**4 is because we are using wavenumbers in cm^-1
        # instead of wavelength in m (see eq 12 of Caldas 2019)

    def sigma(self, wns, vmr):
        """Computes the Rayleigh cross section for the gas.
        This one is faster than sigma_array, but can be used only when vmr values are floats.

        Parameters
        ----------
            wns: array
                array of wavenumbers

            vmr: dict of floats
                Keys are molecule names. Values are the volume mixing ratios

        Returns
        -------
            array of shape (wns.size)
                Rayleigh cross section for the whole gas in m^2/molecule
        """
        res=np.zeros(wns.size)
        wn2 = wns*wns
        wn4 = wn2*wn2
        for mol, x in vmr.items():
            to_add, tmp = self.sigma_mol(mol, wn2, wn4)
            if to_add: res+=x*tmp

        return res

    def sigma_array(self, wns, vmr):
        """Computes the Rayleigh cross section for the gas.

        Parameters
        ----------
            wns: array
                array of wavenumbers

            vmr: dict of arrays
                Keys are molecule names. Values are arrays the volume mixing ratios

        Returns
        -------
            array of shape (vmr.values.size, wns.size)
                Rayleigh cross section for the whole gas in m^2/molecule
        """
        first_mol=True
        wn2 = wns*wns
        wn4 = wn2*wn2
        for mol, x in vmr.items():
            x_array=np.array(x)
            if first_mol:
                res=np.zeros((x_array.size,wns.size))
                first_mol=False
            to_add, tmp = self.sigma_mol(mol, wn2, wn4)
            if to_add: res+=x_array[:,None]*tmp

        return res

    def sigma_mol(self, mol, wn2, wn4):
        """Intermediary function to compute rayleigh for each molecule.

        Parameters
        ----------
            mol: str
                Molecule name.
            wn2, wn4: arrays
                Array of the wavenumber (in cm^-1) to the 2nd and 4th power.
                (To avoid recomputing it each time).
        
        Returns
        -------
            to_add: bool
                Says whether the molecule has been found
                and the contribution needs to be added.
            tmp: array of size self.wns or None
                The cross section for the molecule as a function of wns.
                None if the molecule has not
                been found.

        """
        to_add=True
        tmp=None
        if mol == 'H2':
            #tmp=((8.14E-13)*(wave**(-4.))* \
            #    (1+(1.572E6)*(wave**(-2.))+(1.981E12)*(wave**(-4.))))*1E-4
            tmp=(8.14e-49+1.28e-58*wn2+1.61e-67*wn4)*wn4
        elif mol == 'He':
            #tmp=((5.484E-14)*(wave**(-4.))*(1+(2.44E5)*(wave**(-2.))))*1E-4
            tmp=(5.484e-50+1.338e-60*wn2)*wn4
        elif mol=='N2':
            tmp=self._mult_factor * wn4 * (1.034 + 3.17e-12*wn2) * \
                (6.4982e-5 + 3.0743305e-10/(1.44e10-wn2))**2 * 4./9.
                # 4./9. is approximately ((n+1)/(n**2+2))**2
                # from eq 12 of caldas et al. 2019
        elif mol=='O2':
            tmp=self._mult_factor * wn4 * (1.096 + 1.385e-11*wn2 + 1.448e-20*wn4) * \
                (2.1351e-4 + 0.218567e-10/(0.409e10-wn2))**2 * 4./9.
        elif mol=='CO2':
            tmp=self._mult_factor * wn4 * (1.1364 + 2.53e-11*wn2) * \
                (1.1427e-2 * (5.799e3/(1.6617e10-wn2) + 1.2005e2/(7.960e9-wn2) \
                    + 5.3334/(5.630e9-wn2)))**2 * 4./9.
        else:
            to_add=False
        return to_add, tmp



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Contains classes for atmospheric profiles and their radiative properties.
That's where forward models are computed. 
"""
import numpy as np
from scipy.special import expn
import astropy.units as u
from numba.typed import List
from .gas_mix import Gas_mix
from .util.cst import N_A, PI, RGP, KBOLTZ, RSOL, RJUP, SIG_SB
from .util.radiation import Bnu_integral_num, Bnu, rad_prop_corrk, rad_prop_xsec,\
    Bnu_integral_array, path_integral_corrk, path_integral_xsec
from .util.interp import gauss_legendre
from .util.spectrum import Spectrum


class Atm_profile(object):
    """A class defining an atmospheric PT profile with some global data
    (gravity, etc.)

    The derived class :class:`~exo_k.atm.Atm` handles
    radiative transfer calculations.

    """
    
    def __init__(self, composition={}, psurf=None, ptop=None, logplev=None, tlev=None,
            Tsurf=None, Tstrat=None, grav=None, Rp=None, Mgas=None, rcp=0.28, Nlev=20):
        """Initializes atmospheric profiles

        Parameters
        ----------
            composition: dict
                Keys are molecule names and values the vmr.
                Vmr can be arrays of size Nlev-1 (i.e. the number of layers).
            grav: float
                Planet surface gravity (gravity constant with altitude for now).
            Rp: float of Astropy.unit quantity
                Planet radius. If float, Jupiter radii are assumed.
            rcp: float
                Adiabatic lapse rate for the gas (R/cp)
            Mgas: float, optional
                Molar mass of the gas (kg/mol). If given, overrides the molar mass computed
                from composition.
        
        There are two ways to define the profile.
        Either define:

        * Nlev: int
          Number of level interfaces (Number of layers is Nlev-1)
        * psurf, Tsurf: float
          Surface pressure (Pa) and temperature 
        * ptop: float
          Pressure at the top of the model (Pa) 
        * Tstrat: float
          Stratospheric temperature        

        or:

        * logplev: array
        * tlev: array (same size)
          These will become the pressures (Pa) and temperatures at the level interfaces.
          This will be used to define the surface and top pressures.
          Nlev becomes the size of the arrays. 

        .. warning::
            Layers are counted from the top down (increasing pressure order).
            All methods follow the same convention.
        """
        self.gas_mix=Gas_mix(composition)
        self.rcp=rcp
        if logplev is None:
            self.psurf=psurf
            self.ptop=ptop
            self.Nlev=Nlev
            self.Nlay=Nlev-1
            self.logplev=np.linspace(np.log10(ptop),np.log10(psurf),num=self.Nlev)
            self.plev=10**self.logplev
            self.logplay=(self.logplev[:-1]+self.logplev[1:])*0.5
            self.play=10**self.logplay
            self.set_adiab_profile(Tsurf=Tsurf, Tstrat=Tstrat, rcp=rcp)
        else:
            self.set_logPT_profile(logplev, tlev)
        self.set_Rp(Rp)        
        self.set_grav(grav)
        self.set_Mgas(Mgas)

    def set_logPT_profile(self, log_plev, tlev):
        """Set the logP-T profile of the atmosphere with a new one

        Parameters
        ----------
            log_plev: numpy array
                Log pressure (in Pa) at the level surfaces
            tlev: numpy array (same size)
                temperature at the level surfaces.
        """
        self.logplev=log_plev
        self.plev=10**self.logplev
        self.psurf=self.plev[-1]
        self.ptop=self.plev[0]
        self.tlev=tlev
        self.logplay=(self.logplev[:-1]+self.logplev[1:])*0.5
        self.play=10**self.logplay
        self.tlay=(self.tlev[:-1]+self.tlev[1:])*0.5
        self.Nlev=log_plev.size
        self.Nlay=self.Nlev-1
        self.gas_mix.set_logPT(logp_array=self.logplay, t_array=self.tlay)

    def set_adiab_profile(self, Tsurf=None, Tstrat=None, rcp=0.28):
        """Initializes the logP-T atmospheric profile with an adiabat with index R/cp=rcp

        Parameters
        ----------
            Tsurf: float
                Surface temperature.
            Tstrat: float, optional
                Temperature of the stratosphere. If None is given,
                an isothermal atmosphere with T=Tsurf is returned.
            rcp: float
                R/c_p of the atmosphere
        """
        if Tstrat is None: Tstrat=Tsurf
        self.tlev=Tsurf*(self.plev/self.psurf)**rcp
        self.tlev=np.where(self.tlev<Tstrat,Tstrat,self.tlev)
        self.tlay=(self.tlev[:-1]+self.tlev[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logplay, t_array=self.tlay)

    def set_grav(self, grav=None):
        """Sets the surface gravity of the planet

        Parameters
        ----------
            grav: float
                surface gravity (m/s^2)
        """
        if grav is None: raise RuntimeError('A planet needs a gravity!')
        self.grav=grav
    
    def set_gas(self, composition_dict):
        """Sets the composition of the atmosphere

        Parameters
        ----------
            composition_dict: dictionary
                Keys are molecule names, and values are volume mixing ratios.
                A 'background' value means that the gas will be used to fill up to vmr=1
                If they do not add up to 1 and there is no background gas_mix,
                the rest of the gas_mix is considered transparent.
    
            compute_col_dens: boolean, optional
                If True, the column density per layer of the atmosphere is recomputed.
                This si mostly to save time when we know this will be done later on.
        """
        self.gas_mix.set_composition(composition_dict)
        self.set_Mgas()

    def set_Mgas(self, Mgas=None):
        """Sets the mean molar mass of the atmosphere.

        Parameters
        ----------
            Mgas: float or array of size Nlay
                Mean molar mass (kg/mol).
                If None is give, the mmm is computed from the composition.
        """
        if Mgas is not None:
            self.Mgas=Mgas
        else:
            self.Mgas=self.gas_mix.molar_mass()

    def set_rcp(self,rcp):
        """Sets the adiabatic index of the atmosphere

        Parameters
        ----------
            rcp: float
                R/c_p
        """
        self.rcp=rcp

    def set_Rp(self, Rp):
        """Sets the radius of the planet

        Parameters
        ----------
            Rp: float
                radius of the planet (m)
        """
        if Rp is None:
            self.Rp = None
            return
        if isinstance(Rp,u.quantity.Quantity):
            self.Rp=Rp.to(u.m).value
        else:
            self.Rp=Rp*RJUP

    def set_Rstar(self, Rstar):
        """Sets the radius of the star

        Parameters
        ----------
            Rstar: float
                radius of the star (m)
        """
        if Rstar is None:
            self.Rstar = None
            return
        if isinstance(Rstar,u.quantity.Quantity):
            self.Rstar=Rstar.to(u.m).value
        else:
            self.Rstar=Rstar*RSOL

    def compute_density(self):
        """Computes the number density (m^-3) profile of the atmosphere
        """
        self.density=self.play/(KBOLTZ*self.tlay)

    def compute_layer_col_density(self):
        """Computes the column number density (molecules/m^-2) per layer of the atmosphere
        """
        self.dmass=(self.plev[1:]-self.plev[:-1])/self.grav
        # grav term above should include the altitude effect. 
        #self.dmass[0]=self.plev[1]/self.grav
        # above is just a test extending the atm up to p=0
        if self.Rp is not None:
            self.compute_altitudes()
            self.dmass=self.dmass*(1.+self.zlay/self.Rp)**2
        self.dcol_density=self.dmass*N_A/(self.Mgas)

    def compute_altitudes(self):
        """Compute altitudes of the level surfaces (zlev) and mid layers (zlay).
        """
        H=RGP*self.tlay/(self.grav*self.Mgas)
        dlnP=np.diff(self.logplev)*np.log(10.)
        if self.Rp is None:
            self.dz=H*dlnP
            self.zlay=np.cumsum(self.dz[::-1])
            self.zlev=np.concatenate(([0.],self.zlay))[::-1]
            self.zlay-=0.5*self.dz[::-1]
            self.zlay=self.zlay[::-1]
        else:
            self.zlev=np.zeros_like(self.tlev)
            for i in range(H.size)[::-1]:
                z1=self.zlev[i+1]
                H1=H[i]
                dlnp=dlnP[i]
                self.zlev[i]=z1+( (H1 * (self.Rp + z1)**2 * dlnp) \
                    / (self.Rp**2 + H1 * self.Rp * dlnp + H1 * z1 * dlnp) )
            self.zlay=(self.zlev[:-1]+self.zlev[1:])*0.5
            # the last line is not completely consistent. The integration
            # should be done between layer bottom and mid layer assuming
            # an average T between Tlay and Tlev.
        
    def compute_area(self):
        """Computes the area of the annulus covered by each layer in a transit setup. 
        """
        self.area=PI*(self.Rp+self.zlev[:-1])**2
        self.area[:-1]-=self.area[1:]
        self.area[-1]-=PI*self.Rp**2

    def compute_tangent_path(self):
        """Computes a triangular array of the tangent path length (in m) spent in each layer.
        self.tangent_path[ilay][jlay] is the length that the ray that is tangent to the ilay layer 
        spends in the jlay>=ilay layer (accounting for a factor of 2 due to symmetry)
        """
        if self.Rp is None: raise RuntimeError('Planetary radius should be set')
        self.compute_altitudes()
        self.tangent_path=List()
        # List() is a new numba.typed list to comply with new numba evolution after v0.50
        for ilay in range(self.Nlay):
            z0square=(self.Rp+self.zlay[ilay])**2
            dl=np.sqrt((self.Rp+self.zlev[:ilay+1])**2-z0square)
            dl[:-1]-=dl[1:]
            self.tangent_path.append(2.*dl)


class Atm(Atm_profile):
    """Class based on Atm_profile that handles radiative trasnfer calculations.

    Radiative data are accessed through the :any:`gas_mix.Gas_mix` class.
    """
    
    def __init__(self, k_database=None, cia_database=None,
        wn_range=None, wl_range=None, **kwargs):
        """Initialization method that calls Atm_Profile().__init__() and links
        to Kdatabase and other radiative data. 
        """
        super().__init__(**kwargs)
        self.set_k_database(k_database)
        self.set_cia_database(cia_database)
        self.set_spectral_range(wn_range=wn_range, wl_range=wl_range)

    def set_k_database(self, k_database=None):
        """Change the radiative database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_k_database` for details.

        Parameters
        ----------
            k_database: :class:`Kdatabase` object
                New Kdatabase to use.
        """
        self.gas_mix.set_k_database(k_database=k_database)
        self.kdatabase=self.gas_mix.kdatabase
        self.Ng=self.gas_mix.Ng
        # to know whether we are dealing with corr-k or not and access some attributes. 

    def set_cia_database(self, cia_database=None):
        """Change the CIA database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_cia_database` for details.

        Parameters
        ----------
            cia_database: :class:`CIAdatabase` object
                New CIAdatabase to use.
        """
        self.gas_mix.set_cia_database(cia_database=cia_database)

    def set_spectral_range(self, wn_range=None, wl_range=None):
        """Sets the spectral range in which computations will be done by specifying
        either the wavenumber (in cm^-1) or the wavelength (in micron) range.

        See :any:`gas_mix.Gas_mix.set_spectral_range` for details.
        """
        self.gas_mix.set_spectral_range(wn_range=wn_range, wl_range=wl_range)

    def opacity(self, **kwargs):
        """Computes the opacity of each of the layers.

        See :any:`gas_mix.Gas_mix.cross_section` for details.
        """
        self.kdata = self.gas_mix.cross_section(**kwargs)
        self.Nw=self.gas_mix.Nw
        self.wns=self.gas_mix.wns
        self.wnedges=self.gas_mix.wnedges

    def emission_spectrum(self, integral=True, mu0=0.5, mu_quad_order=None, **kwargs):
        """Computes the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            mu_quad_order: int
                If an integer is given, the emission intensity is computed
                for a number of angles and integrated following a gauss legendre
                quadrature rule of order `mu_quad_order`.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        if mu_quad_order is not None:
            # if we want quadrature, use the more general method.
            return self.emission_spectrum_quad(integral=integral,
                mu_quad_order=mu_quad_order, **kwargs)

        self.opacity(rayleigh=False, **kwargs)
        if integral:
            #JL2020 What is below is slightly faster for very large resolutions
            #dw=np.diff(self.wnedges)
            #piBatm=np.empty((self.Nlev,self.Nw))
            #for ii in range(self.Nlev):
            #    piBatm[ii]=PI*Bnu_integral_num(self.wnedges,self.tlev[ii])/dw
            #JL2020 What is below is much faster for moderate to low resolutions
            piBatm=PI*Bnu_integral_array(self.wnedges,self.tlev,self.Nw,self.Nlev) \
                /np.diff(self.wnedges)
        else:
            piBatm=PI*Bnu(self.wns[None,:],self.tlev[:,None])
        self.SurfFlux=piBatm[-1]
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau,dtau=rad_prop_xsec(self.dcol_density,self.kdata,mu0)
        else:
            self.tau,dtau=rad_prop_corrk(self.dcol_density,self.kdata,mu0)
            weights=self.kdatabase.weights
        expdtau=np.exp(-dtau)
        expdtauminone=np.where(dtau<1.e-14,-dtau,expdtau-1.)
        # careful: due to numerical limitations, 
        # the limited development of Exp(-dtau)-1 needs to be used for small values of dtau
        exptau=np.exp(-self.tau)
        if self.Ng is None:
            timesBatmTop=(-expdtau-expdtauminone/dtau)*exptau[:-1]
            timesBatmBottom=(1.+expdtauminone/dtau)*exptau[:-1]
            timesBatmBottom[-1]=timesBatmBottom[-1]+exptau[-1]
        else:
            timesBatmTop=np.sum((-expdtau-expdtauminone/dtau)*exptau[:-1]*weights,axis=2)
            timesBatmBottom=np.sum((1.+expdtauminone/dtau)*exptau[:-1]*weights,axis=2)
            timesBatmBottom[-1]=timesBatmBottom[-1]+np.sum(exptau[-1]*weights,axis=1)
        IpTop=np.sum(piBatm[:-1]*timesBatmTop+piBatm[1:]*timesBatmBottom,axis=0)

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_quad(self, integral=True, mu_quad_order=3, **kwargs):
        """Computes the emission flux at the top of the atmosphere (in W/m^2/cm^-1)
        using gauss legendre qudrature of order `mu_quad_order`

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
                  
        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.opacity(rayleigh=False, **kwargs)
        if integral:
            piBatm=2.*PI*Bnu_integral_array(self.wnedges,self.tlev,self.Nw,self.Nlev) \
                /np.diff(self.wnedges)
        else:
            piBatm=2.*PI*Bnu(self.wns[None,:],self.tlev[:,None])
        self.SurfFlux=piBatm[-1]

        IpTop=np.zeros(self.kdata.shape[1])
        mu_w, mu_a, _ = gauss_legendre(mu_quad_order)
        mu_w = mu_w * mu_a # takes care of the mu factor in last integral => int(mu I d mu)
        self.compute_layer_col_density()
        for ii, mu0 in enumerate(mu_a):
            if self.Ng is None:
                self.tau,dtau=rad_prop_xsec(self.dcol_density,self.kdata,mu0)
            else:
                self.tau,dtau=rad_prop_corrk(self.dcol_density,self.kdata,mu0)
                weights=self.kdatabase.weights
            expdtau=np.exp(-dtau)
            expdtauminone=np.where(dtau<1.e-14,-dtau,expdtau-1.)
            exptau=np.exp(-self.tau)
            if self.Ng is None:
                timesBatmTop=(-expdtau-expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom=(1.+expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom[-1]=timesBatmBottom[-1]+exptau[-1]
            else:
                timesBatmTop=np.sum((-expdtau-expdtauminone/dtau)*exptau[:-1]*weights,axis=2)
                timesBatmBottom=np.sum((1.+expdtauminone/dtau)*exptau[:-1]*weights,axis=2)
                timesBatmBottom[-1]=timesBatmBottom[-1]+np.sum(exptau[-1]*weights,axis=1)
            IpTop+=np.sum(piBatm[:-1]*timesBatmTop+piBatm[1:]*timesBatmBottom,axis=0)*mu_w[ii]

        return Spectrum(IpTop,self.wns,self.wnedges)

    def exp_minus_tau(self):
        """Sums Exp(-tau) over gauss points
        """
        weights=self.kdatabase.weights
        return np.sum(np.exp(-self.tau[1:])*weights,axis=2)

    def exp_minus_tau_g(self, g_index):
        """Sums Exp(-tau) over gauss points
        """
        return np.exp(-self.tau[1:,:,g_index])

    def surf_bb(self, integral=True):
        """Computes the surface black body flux (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
        Returns
        -------
            Spectrum object
                Spectral flux at the surface (in W/m^2/cm^-1)
        """
        if integral:
            piBatm=PI*Bnu_integral_num(self.wnedges,self.tlev[-1])/np.diff(self.wnedges)
        else:
            piBatm=PI*Bnu(self.wns[:],self.tlev[-1])
        return Spectrum(piBatm,self.wns,self.wnedges)

    def top_bb(self, integral=True):
        """Computes the top of atmosphere black body flux (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                If true, the black body is integrated within each wavenumber bin.
                If not, only the central value is used.
                    False is faster and should be ok for small bins,
                    but True is the correct version. 
        Returns
        -------
            Spectrum object
                Spectral flux of a bb at the temperature at the top of atmosphere (in W/m^2/cm^-1)
        """
        if integral:
            piBatm=PI*Bnu_integral_num(self.wnedges,self.tlev[0])/np.diff(self.wnedges)
        else:
            piBatm=PI*Bnu(self.wns[:],self.tlev[0])
        return Spectrum(piBatm,self.wns,self.wnedges)

    def transmittance_profile(self, **kwargs):
        """Computes the transmittance profile of an atmosphere,
        i.e. Exp(-tau) for each layer of the model.
        Real work done in the numbafied function path_integral_corrk/xsec
        depending on the type of data.
        """
        self.opacity(**kwargs)
        self.compute_tangent_path()
        self.compute_density()
        if self.Ng is not None:
            weights=self.kdatabase.weights
            transmittance=path_integral_corrk( \
                self.Nlay,self.Nw,self.Ng,self.tangent_path,self.density,self.kdata,weights)
        else:
            transmittance=path_integral_xsec( \
                self.Nlay,self.Nw,self.tangent_path,self.density,self.kdata)
        return transmittance

    def transmission_spectrum(self, normalized=False, Rstar=None, **kwargs):
        """Computes the transmission spectrum of the atmosphere.
        In general (see options below), the code returns the transit depth:

        delta_nu=(pi*Rp^2+alpha_nu)/(pi*Rstar^2).

        where

        alpha_nu=2*pi*Int_0^zmax (Rp+z)*(1-exp(-tau_nu(z))) dz
        
        Parameters
        ----------
            Rstar: float, optional
                Radius of the host star. Does not need to be given here if
                as already been specified as an attribute of the self.Atm object.
                If specified, the result is the transit depth:
                delta_nu=(pi*Rp^2+alpha_nu)/(pi*Rstar^2).
            normalized: boolean, optional
                Used only if self.Rstar and Rstar are None.
                * If true, the result is normalized to the planetary radius:
                  delta_nu=1+alpha_nu/(pi*Rp^2).
                * If False, delta_sigma=1+(h_sigma/R_planet)^2.
                  delta_nu=pi*Rp^2+alpha_nu.

        """
        self.set_Rstar(Rstar)
        transmittance=self.transmittance_profile(**kwargs)
        self.compute_area()
        res=Spectrum((np.dot(self.area,(1.-transmittance))),self.wns,self.wnedges)
        if self.Rstar is not None:
            return (res+(PI*self.Rp**2))/(PI*self.Rstar**2)
        elif normalized:
            return res/(PI*self.Rp**2)+1
        else:
            return res+(PI*self.Rp**2)

    def heating_rate(self, Fin=1., Tstar=5570., szangle=60., **kwargs):
        """Computes the heating rate in the atmosphere
        """
        mu0=np.cos(szangle*PI/180.)
        self.opacity(rayleigh=False, **kwargs)
        Fstar=Fin*PI*Bnu_integral_array(self.wnedges,List([Tstar]),self.Nw,1)/(SIG_SB*Tstar**4)
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau, _ =rad_prop_xsec(self.dcol_density,self.kdata,mu0)
            #the second returned variable is ignored
        else:
            self.tau, _ =rad_prop_corrk(self.dcol_density,self.kdata,mu0)
            weights=self.kdatabase.weights
        exptau=np.exp(-self.tau)
        if self.Ng is None:
            tmp_heat_rate=Fstar[0,:]*exptau
            tmp_heat_rate[:-1]-=tmp_heat_rate[1:]
            heat_rate=tmp_heat_rate[:-1]
        else:
            tmp_heat_rate=Fstar[0,:,None]*exptau
            tmp_heat_rate[:-1]-=tmp_heat_rate[1:]
            heat_rate=np.sum(tmp_heat_rate[:-1]*weights,axis=2)
        heat_rate=np.sum(heat_rate,axis=1)
        heat_rate=heat_rate*N_A*self.rcp/(self.dcol_density*RGP)
        return heat_rate

############### unvalidated functions #########################

    def emission_spectrum_exp_integral(self, integral=True, **kwargs):
        """Computes the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        .. warning::
            This method uses a formulation with exponential integrals that
            is not yet perceived as accurate enough by the author. It is left for testing only. 

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        #if self.dtau is None:
        #    self.rad_prop(**kwargs)
        self.opacity(rayleigh=False, **kwargs)
        if integral:
            piBatm=2*PI*Bnu_integral_array(self.wnedges,self.tlev,self.Nw,self.Nlev) \
                /np.diff(self.wnedges)
        else:
            piBatm=2*PI*Bnu(self.wns[None,:],self.tlev[:,None])
        self.SurfFlux=piBatm[-1]
        mu_zenith=1.
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau,dtau=rad_prop_xsec(self.dcol_density, self.kdata, mu_zenith)
        else:
            self.tau,dtau=rad_prop_corrk(self.dcol_density, self.kdata, mu_zenith)
            weights=self.kdatabase.weights
        factor=expn(2, self.tau[1:])*dtau
        if self.Ng is None:
            timesBatmTop=factor*piBatm[:-1]
            surf_contrib=piBatm[-1]*expn(3, self.tau[-1])
        else:
            timesBatmTop=piBatm[:-1]*np.sum(factor*weights,axis=2)
            surf_contrib=piBatm[-1]*np.sum(expn(3, self.tau[-1])*weights,axis=1)
        IpTop=np.sum(timesBatmTop,axis=0)+surf_contrib
        return Spectrum(IpTop, self.wns, self.wnedges)




# -*- coding: utf-8 -*-
"""
Created on Jun 20 2019

@author: jeremy leconte

Library of useful functions for radiative transfer calculations
"""
import numpy as np
import numba
import scipy.integrate as integrate
from exo_k.util.cst import PI,PLANCK,C_LUM,KBOLTZ,SIG_SB

PLANCK_CST1MKS=2.*PLANCK*C_LUM**2
PLANCK_CST1=1.e2*PLANCK_CST1MKS
PLANCK_CST2=PLANCK*C_LUM/(KBOLTZ)
PLANCK_CST1_lamb=1.e-6*2.*PLANCK*C_LUM**2




@numba.jit(nopython=True,fastmath=True)
#@numba.vectorize([float64(float64,float64)],fastmath=True)
def Bnu(nu, T):
    """Computes the Planck law in wavenumber domain.

    Parameters
    ----------
        nu: float
            The wavenumber in cm^-1. 
        T: float
            Temperature (K).

    Returns
    -------
        float
            Bnu in W/m^2/str/cm^-1 (not SI units).

    sigma*T^4=Int(Bnu dnu) where nu is expressed in cm^-1
    """
    sigma=nu*1.e2
    return PLANCK_CST1*sigma**3/(np.exp(PLANCK_CST2*sigma/T)-1.)


def Bnu_integral(nu_edges, T):
    """Computes the integral of the Planck function in wavenumber bins.
    
    Parameters
    ----------
        nu_edges: array
            Edges of the wavenumber bins in cm^-1.
        T: float
            Temperature (K).

    Returns
    -------
        array
            Bnu_integral in W/m^2/str.

    sigma*T^4/PI=sum of Bnu_integral
    """
    nu_edges=np.array(nu_edges)
    res=np.empty(nu_edges.size-1)
    for ii in range(nu_edges.size-1):
        res[ii]=integrate.quad(lambda nu: Bnu(nu,T),nu_edges[ii],nu_edges[ii+1])[0]
    return res

@numba.jit(nopython=True, fastmath=True)
def Bnu_integral_num(nu_edges, T, n=30):
    """Computes the integral of the Planck function in wavenumber bins.
    
    Parameters
    ----------
        nu_edges: array
            Edges of the wavenumber bins in cm^-1.
        T: float
            Temperature (K).
        n: int
            number of terms to take into account in the black body calculation.
    
    Returns
    -------
        Array of shape (T_array.size,nu_edges.size-1)
            Bnu_integral_num: Integral of the source function at 
            temperatures T_array inside the bins in W/m^2/str
    
    sigma*T^4=sum of Bnu_integral_num
    """
    kp=PLANCK_CST2/T
    res=np.zeros(nu_edges.size)
    Nn=nu_edges.size
    for ii in range(Nn):
        kpnu=kp*nu_edges[ii]*1.e2
        for iteration in range(1,n+1):
            kpnun=kpnu*iteration
            #res[ii]+=np.exp(-kpnun)*(6+6*kpnun+3*kpnun**2+kpnun**3)/iteration**4
            res[ii]+=np.exp(-kpnun)*(6.+6.*kpnun+3.*kpnun**2+kpnun**3)/iteration**4
    for ii in range(Nn-1):
        res[ii]-=res[ii+1]

    return res[:-1]*PLANCK_CST1MKS/kp**4

@numba.jit(nopython=True, fastmath=True)
def Bnu_integral_array(nu_edges, T_array, Nw, Nt, n=30):
    """Computes the integral of the Planck function in wavenumber bins.
    
    Parameters
    ----------
        nu_edges: array
            Edges of the wavenumber bins in cm^-1.
        T: array
            Array of temperature (K).
        Nw: int
            Number of spectral bins.
        Nt: int
            Number of temperatures.
        n: int, optional
            number of terms to take into account in the black body calculation.
    
    Returns
    -------
        Array of shape (T_array.size,nu_edges.size-1)
            Bnu_integral_num: Integral of the source function at
            temperatures T_array inside the bins in W/m^2/str
    
    sigma*T^4=sum of Bnu_integral_num
    """
    res=np.empty((Nt,Nw))
    for iT in range(Nt):
        kp=PLANCK_CST2/T_array[iT]
        mult=PLANCK_CST1MKS/kp**4
        kpnu=kp*nu_edges[0]*1.e2
        tmp=0.
        for iteration in range(1,n+1):
            kpnun=kpnu*iteration
#            tmp+=np.exp(-kpnun)*(6+6*kpnun+3*kpnun**2+kpnun**3)/iteration**4
            tmp+=np.exp(-kpnun)*(6.+6.*kpnun+3.*kpnun**2+kpnun**3)/iteration**4
        for ii in range(1,Nw+1):
            tmp2=0.
            kpnu=kp*nu_edges[ii]*1.e2
            for iteration in range(1,n+1):
                kpnun=kpnu*iteration
#                tmp2+=np.exp(-kpnun)*(6+6*kpnun+3*kpnun**2+kpnun**3)/iteration**4
                tmp2+=np.exp(-kpnun)*(6.+6.*kpnun+3.*kpnun**2+kpnun**3)/iteration**4
            res[iT,ii-1]=(tmp-tmp2)*mult
            tmp=tmp2
    return res

def Bnu_integral_old(nu_edges, T):
    """Deprecated.
    
    Computes the integral of the Planck function in wavenumber bins.
    nu_edges is an array of the edges of the wavenumber bins in cm^-1.
    
    sigma*T^4=sum of Bnu_integral

    Bnu_integral is in W/m^2/str
    """
    a1=1.e8*2.*PLANCK*C_LUM**2
    a2=1.e2*PLANCK*C_LUM/(KBOLTZ*T)
    res=np.empty(nu_edges.size-1)
    for ii in range(nu_edges.size-1):
        res[ii]=integrate.quad(lambda nu: nu**3/(np.exp(nu*a2)-1.),nu_edges[ii],nu_edges[ii+1])[0]
    return res*a1


@numba.jit(nopython=True)
def Bmicron(lamb, T):
    """Computes the Planck law in wavelength domain.

    Parameters
    ----------
        lamb: float
            The wavelength in microns. 
        T: float
            Temperature (K).
 
    Returns:
        float
            Bmicron in W/m^2/str/micron (not SI units).

    sigma*T^4=Int(Bmicron d lamb) where lamb is expressed in microns
    """
    lambda_si=lamb*1.e-6
    return PLANCK_CST1_lamb/(lambda_si**5   \
        *(np.exp(PLANCK_CST2/(lambda_si*T))-1.))


@numba.jit(nopython=True)
def BBnu_Stellar_Spectrum(nu, T, Flux):
    """Computes the outgoing spectral flux of a black body at temperature T so that its 
    bolometric flux is equal to the input 'Flux'

    Parameters
    ----------
        nu: float
            The wavenumber in cm^-1. 
        T: float
            Temperature (K).
        Flux: float
            bolometric flux for rescaling.
    
    Returns
    -------
        float
            BBnu_Stellar_Spectrum is in W/m^2/cm^-1 (not SI units).
    
    sigma*T^4=Int(Bnu dnu) where nu is expressed in cm^-1
    """
    Scaling_Factor=PI/(SIG_SB*T**4)*Flux
    return Scaling_Factor*Bnu(nu,T)
    
@numba.jit(nopython=True)
def BBmicron_Stellar_Spectrum(lamb, T, Flux):
    """Computes the outgoing spectral flux of a black body at temperature T so that its 
    bolometric flux is equal to the input 'Flux'

    Parameters
    ----------
        lamb: float
            The wavelength in microns. 
        T: float
            Temperature (K).
        Flux: float
            bolometric flux for rescaling.
    
    Returns
    -------
        float
            BBmicron_Stellar_Spectrum is in W/m^2/micron (not SI units).
            
    sigma*T^4=Int(Bmicron d lamb) where lamb is expressed in microns.
    
    """
    Scaling_Factor=PI/(SIG_SB*T**4)*Flux
    return Scaling_Factor*Bmicron(lamb,T)

def wavelength_grid_R(lambda_min, lambda_max, R):
    """Creates a wavelength grid starting from lambda_min and ending at lambda_max
    with a resolution of R (roughly).

    Parameters
    ----------
        lambda_min/lambda_max: int or float
            Min/max wavelength (in micron).
        R: int or float
            Resolution.

    Returns
    -------
        array
            Grid of wavelength.
    """
    return np.append(np.exp(np.arange(np.log(lambda_min),np.log(lambda_max),1./R)),lambda_max)

def wavenumber_grid_R(nu_min, nu_max, R):
    """Creates a wavenumber grid starting from nu_min and ending at nu_max
    with a resolution of R (roughly).

    Parameters
    ----------
        nu_min/nu_max: int or float
            Min/max wavenumber (in cm^-1).
        R: int or float
            Resolution.

    Returns
    -------
        array
            Grid of wavenumber.
    """
    return np.append(np.exp(np.arange(np.log(nu_min),np.log(nu_max),1./R)),nu_max)

@numba.njit
def rad_prop_corrk(dcol_density, opacity_prof, mu0):
    """Computes the optical depth of each of the layers
    for the opacity given for every wavelength and g point.

    Parameters
    ----------
        dcol_density: array
            Column density per layer (molecules/m^2/layer).
        opacity_prof: array
            effective cross section per molecule (m^2/molecule).
        mu0: cosine of the equivalent zenith angle for the rays.
            e.g. Cos(60deg)=0.5 is chosen (See, for example, chap 4 of Pierrehumbert 2010)

    Returns
    -------
        tau: Array
            cumulative optical depth from the top (with zeros at the top of the first layer)
        dtau: Array
            optical depth of each layer for each wavenumber.
   """
    Nlev,Nw,Ng=opacity_prof.shape
    OvMu=1./mu0
    dtau=np.empty((Nlev,Nw,Ng))
    tau=np.zeros((Nlev+1,Nw,Ng))
    for lev in range(Nlev):
        for iW in range(Nw):
            for ig in range(Ng):
                dtau_tmp=dcol_density[lev]*opacity_prof[lev,iW,ig]*OvMu
                dtau[lev,iW,ig]=dtau_tmp
                tau[lev+1,iW,ig]=tau[lev,iW,ig]+dtau_tmp
    return tau, dtau

@numba.njit
def rad_prop_xsec(dcol_density, opacity_prof, mu0):
    """Computes the optical depth of each of the layers
    for the opacity given for every wavelength.

    Parameters
    ----------
        dcol_density: array
            Column density per layer (molecules/m^2/layer).
        opacity_prof: array
            effective cross section per molecule (m^2/molecule).
        mu0: cosine of the equivalent zenith angle for the rays.
            e.g. Cos(60deg)=0.5 is chosen (See, for example, chap 4 of Pierrehumbert 2010)

    Returns
    -------
        tau: Array
            cumulative optical depth from the top (with zeros at the top of the first layer)
        dtau: Array
            optical depth of each layer for each wavenumber.
    """
    Nlev,Nw=opacity_prof.shape
    OvMu=1./mu0
    dtau=np.empty((Nlev,Nw))
    tau=np.zeros((Nlev+1,Nw))
    for lev in range(Nlev):
        for iW in range(Nw):
            dtau_tmp=dcol_density[lev]*opacity_prof[lev,iW]*OvMu
            dtau[lev,iW]=dtau_tmp
            tau[lev+1,iW]=tau[lev,iW]+dtau_tmp
    return tau,dtau

@numba.njit(nogil=True,fastmath=True)
def path_integral_corrk(Nlay, Nw, Ng, tangent_path, density_prof, opacity_prof, weights):
    """Computes the transmittance for each layer of the atmosphere for k-coefficients.

    Parameters
    ----------
        Nlay: int
            Number of layers.
        Nw: int
            Number of wavenumber bins.
        Ng: int
            Number of gauss points.
        tangent_path: List of arrays
            Triangular array of tangent path for each layer.
        density_prof: array
            Array with the number density of molecules in the atmosphere.
        opacity_prof: array
            Effective cross section of the atmsophere for each (layer,wavenumber,g-point).
        weights: array
            Weights for the quadrature in g-space.

    Returns
    -------
        transmittance: array
            Transmittance for each layer and wavenumber bin (Exp(-tau_sigma)).
    """
    exp_min_tau=np.zeros((Nlay,Nw,Ng))
    transmittance=np.zeros((Nlay,Nw))
    for ilay in range(Nlay):
        for jlay in range(ilay+1):
            dm=tangent_path[ilay][jlay]*density_prof[jlay]
            for iW in range(Nw):
                for ig in range(Ng):
                    exp_min_tau[ilay,iW,ig]+=dm*opacity_prof[jlay,iW,ig]
    exp_min_tau=np.exp(-exp_min_tau)
    for ilay in range(Nlay):
        for iW in range(Nw):
            trans=0.
            for ig in range(Ng):
                trans+=exp_min_tau[ilay,iW,ig]*weights[ig]
            transmittance[ilay,iW]=trans
    return transmittance


@numba.njit(nogil=True,fastmath=True)
def path_integral_xsec(Nlay, Nw, tangent_path, density_prof, opacity_prof):
    """Computes the transmittance for each layer of the atmosphere from cross sections.

    Parameters
    ----------
        Nlay: int
            Number of layers.
        Nw: int
            Number of wavenumber bins.
        tangent_path: List of arrays
            Triangular array of tangent path for each layer.
        density_prof: array
            Array with the number density of molecules in the atmosphere.
        opacity_prof: array
            Effective cross section of the atmsophere for each (layer,wavenumber).

    Returns
    -------
        transmittance: array
            Transmittance for each layer and wavenumber bin (Exp(-tau_sigma)).
    """
    transmittance=np.zeros((Nlay,Nw))
    for ilay in range(Nlay):
        for jlay in range(ilay+1):
            dm=tangent_path[ilay][jlay]*density_prof[jlay]
            for iW in range(Nw):
                transmittance[ilay,iW]+=dm*opacity_prof[jlay,iW]
    transmittance=np.exp(-transmittance)
    return transmittance




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from .util.molar_mass import Molar_mass
from .util.interp import RandOverlap_2_kdata_prof, rm_molec
from .rayleigh import Rayleigh

class Gas_mix(object):
    """Dict-like class to handle gas composition (with background gas) and molar mass.

    If `logp_array`, `t_array`, and radiative databases are provided, :any:`cross_section`
    can be used to compute the opacity of the gas
    """

    def __init__(self, composition={}, bg_gas=None, logp_array=None, t_array=None,
        k_database=None, cia_database=None):
        """__init_ Instantiates
        a Gas_mix object and computes the vmr of the 'background' gas.
        """
        self.Narray=None
        self._wn_range=None
        self.iw_min=None
        self.iw_max=None
        self.wns=None
        self.wnedges=None
        self.Nw=None

        self.set_composition(composition, bg_gas=bg_gas)
        self.set_logPT(logp_array=logp_array, t_array=t_array)
        self.set_k_database(k_database)
        self.set_cia_database(cia_database)


    def set_composition(self, composition, bg_gas=None):
        """Reset composition and computes the vmr of the 'background' gas.

        Parameters
        ----------
            composition: dict
                Keys are molecule names. Values are vmr or arrays of vmr.
            bg_gas: str
                Name of the background molecule. 
                If None, it is inferred from the molecule for which vmr='background'.
        """
        self.composition=dict()
        for mol,vmr in composition.items():
            if isinstance(vmr, list):
                self.composition[mol]=np.array(vmr)
            else:
                self.composition[mol]=vmr
        self.bg_gas=bg_gas
        if (True in [isinstance(val,str) for val in self.composition.values()]) \
                or (self.bg_gas is not None):
            self.composition['inactive_gas']=0.
        else:
            self.composition['inactive_gas']='background'
            self.bg_gas='inactive_gas'
        self.get_background_vmr()

    def set_logPT(self, logp_array=None, t_array=None):
        """Sets the pressure (in Pa) and temperature fields.
        """
        if logp_array is not None:
            if hasattr(logp_array, "__len__"):
                self.logp_array=np.array(logp_array, dtype=float)
            else:
                self.logp_array=np.array([logp_array], dtype=float)

            if hasattr(t_array, "__len__"):
                self.t_array=np.array(t_array)
            else:
                self.t_array=np.array([t_array])

            if self.logp_array.size != self.t_array.size:
                raise TypeError(\
                    'logp_array and t_array should be 1d lists or arrays of the same size')
            self.Narray=self.logp_array.size

    def get_background_vmr(self):
        """Computes the volume mixing ratio of the background gas in a mix
        Uses the fact that self.composition is a dictionary
        with the name of the molecules as keys and the vmr as values.
        vmr are either float or arrays.
        
        At this stage, the background gas should be identified by `self.bg_gas`,
        and its Vol. Mix. Ratio will be updated in the composition dict. 
        """
        other_vmr=0.
        if self.bg_gas is None: # should never happen
            for mol,vmr in self.composition.items():
                if isinstance(vmr,str):
                    self.bg_gas=mol
                    continue
                try:
                    other_vmr+=vmr
                except ValueError:
                    raise TypeError('Incompatible shapes in Gas_mix arrays.')
        else:
            for mol,vmr in self.composition.items():
                if mol==self.bg_gas:
                    continue
                try:
                    other_vmr+=vmr
                except ValueError:
                    raise TypeError('Incompatible shapes in Gas_mix arrays.')
        if np.amax(other_vmr)>1.:
            print("""Carefull: the sum of the vmr of your gas components is > 1.
            If there is a background gas, its vmr will become negative.
            I hope you know what you are doing.""")
        self.composition[self.bg_gas]=1.-other_vmr

    def molar_mass(self):
        """Computes the molar mass of a mix of gases

        Returns
        -------
            float or array:
                Molar mass of the active gases in kg/mol
        """
        mol_mass_active_gases=0.
        vmr_active_gases=0.

        for mol,vmr in self.composition.items():
            if mol!='inactive_gas':
                Mmol=Molar_mass().fetch(mol)
                mol_mass_active_gases+=vmr*Mmol
                vmr_active_gases+=vmr
        mol_mass=mol_mass_active_gases/vmr_active_gases     
        return mol_mass

    def get_vmr_array(self, sh=None):
        """Returns a dictionary with an array of vol. mix. ratios for each species. 

        Parameters
        ----------
            sh: set or list
                shape of the array wanted if all the vmr are floats.
                If some are already arrays, check whether the shape is the correct one. 

        Returns
        -------
            res: dict
                A dictionary with the an array of vmr per species.
            cst_array: boolean
                Is True if all the values in the arrays are constant.
        """
        res=dict()
        cst_array=True
        for mol,vmr in self.composition.items():
            if isinstance(vmr,(float,int)):
                res[mol]=np.ones(sh)*vmr
            else:
                cst_array=False
                res[mol]=np.array(vmr)
                if not np.array_equal(res[mol].shape, sh):
                    raise RuntimeError('Wrong shape in get_vmr_array')
        return res, cst_array

    def set_k_database(self, k_database=None):
        """Change the radiative database attached to the current instance of Gas_mix

        Parameters
        ----------
            k_database: :class:`~exo_k.kdatabase.Kdatabase` object
                New Kdatabase to use.
        """
        self.kdatabase=k_database
        if self.kdatabase is None:
            self.Ng=None
        else:
            self.Ng=self.kdatabase.Ng
            if self.kdatabase.p_unit != 'Pa' or rm_molec(self.kdatabase.kdata_unit) != 'm^2':
                print("""You're being Bad!!! You are trying *NOT* to use MKS units!!!
                You can convert to mks using convert_to_mks.
                More generally, you can specify exo_k.Settings().set_mks(True) 
                to force automatic conversion to mks of all newly loaded data.
                You will have to reload all your data though.
                (A good thing it does not take so long). """)
                raise RuntimeError("Bad units in the Kdatabase used with Gas_mix.")
            if (not self.kdatabase.consolidated_p_unit) \
                or (not self.kdatabase.consolidated_kdata_unit):
                raise RuntimeError( \
                    """All tables in the database should have the same units to proceed.
                    You should probably use convert_to_mks().""")

    def set_cia_database(self, cia_database=None):
        """Change the CIA database attached to the current instance of Gas_mix

        Parameters
        ----------
            cia_database: :class:`~exo_k.ciadatabase.CIAdatabase` object
                New CIAdatabase to use.
        """
        self.cia_database=cia_database
        if self.cia_database is not None:
            if self.cia_database.abs_coeff_unit != 'm^5':
                print("""You're being Bad!!! You are trying *NOT* to use MKS units!!!
                You can convert to mks using convert_to_mks.
                More generally, you can specify exo_k.Settings().set_mks(True) 
                to force automatic conversion to mks of all newly loaded data.
                You will have to reload all your data though.
                (A good thing it does not take so long). """)
                raise RuntimeError("Bad units in the CIAdatabase used with Gas_mix.")

    def set_spectral_range(self, wn_range=None, wl_range=None):
        """Sets the spectral range in which computations will be done by specifying
        either the wavenumber or the wavelength range.

        Parameters
        ----------
            wn_range: list or array of size 2
                Minimum and maximum wavenumber (in cm^-1).
            wl_range: list or array of size 2
                Minimum and maximum wavelength (in micron)
        """
        if wl_range is not None:
            if wn_range is not None:
                print('Cannot specify both wl and wn range!')
                raise RuntimeError()
            else:
                self._wn_range=np.sort(10000./np.array(wl_range))
        else:
            if wn_range is not None:
                self._wn_range=np.sort(np.array(wn_range))

    def compute_wn_range_indices(self):
        """Compute the starting and ending indices to be used for current wn_range
        """
        if self._wn_range is None:
            self.iw_min=0
            self.iw_max=self.kdatabase.Nw
        else:
            self.iw_min, self.iw_max = np.where((self.kdatabase.wnedges > self._wn_range[0]) \
                & (self.kdatabase.wnedges <= self._wn_range[1]))[0][[0,-1]]
            # to be consistent with interpolate_kdata

    def cross_section(self, composition=None, logp_array=None, t_array=None,
            wl_range=None, wn_range=None, rayleigh=True,
            write=0, random_overlap=False, **kwargs):
        """Computes the cross section (m^2/total number of molecule) for the mix
        at each of the logPT points as a function of wavenumber (and possibly g point).

        Parameters
        ----------
            wl_range: array or list of two values, optional
                Wavelength range to cover.
            wn_range: array or list of two values, optional
                Wavenumber range to cover.
            rayleigh: boolean, optional
                Whether to compute rayleigh scattering.
            random_overlap: boolean, optional
                Whether Ktable opacities are added linearly (False),
                or using random overlap method (True).

        Returns
        -------
            kdata_array: array
                Cross section array of shape (layer number, Nw (, Ng if corrk)).

        After every computation, the following variables are updated to account for any possible
        change in spectral range:

          * self.Nw, Number of wavenumber bins
          * self.wns, Wavenumber array
          * self.wnedges, Wavenumber of the edges of the bins
        """
        if self.kdatabase is None: raise RuntimeError("""k_database not provided. 
        Use the k_database keyword during initialization or use the set_k_database method.""")
        if not self.kdatabase.consolidated_wn_grid: raise RuntimeError("""
            All tables in the database should have the same wavenumber grid to proceed.
            You should probably use bin_down().""")
        if self.cia_database is not None and \
            (not np.array_equal(self.cia_database.wns,self.kdatabase.wns)):
            raise RuntimeError("""CIAdatabase not sampled on the right wavenumber grid.
            You should probably run something like CIAdatabase.sample(Kdatabase.wns).""")
        
        self.set_logPT(logp_array=logp_array, t_array=t_array)
        if self.Narray is None:
            raise RuntimeError('You must prescribe logP (in Pa) and T arrays first.')

        if composition is not None:
            self.set_composition(composition)
        molecs=self.composition.keys()
        mol_to_be_done=list(set(molecs).intersection(self.kdatabase.molecules))
        if all(elem in self.kdatabase.molecules for elem in molecs):
            if write>3 : print("""I have all the molecules present in the atmosphere
              in ktables provided:""")
        else:
            if write>3 : print("""Some missing molecules in my database,
             I ll compute opacites with the available ones:""")
        if write>3 : print(mol_to_be_done)

        self.set_spectral_range(wl_range=wl_range, wn_range=wn_range)
        self.compute_wn_range_indices()
        self.wnedges=np.copy(self.kdatabase.wnedges[self.iw_min:self.iw_max+1])
        self.wns=np.copy(self.kdatabase.wns[self.iw_min:self.iw_max])
        self.Nw=self.wns.size

        vmr_prof, cst_prof=self.get_vmr_array((self.Narray,))
        first_mol=True
        for mol in mol_to_be_done:
            tmp_kdata=self.kdatabase[mol].interpolate_kdata(logp_array=self.logp_array,
                t_array=self.t_array, x_array=vmr_prof[mol], wngrid_limit=self._wn_range,
                **kwargs)
            if first_mol:
                kdata_array=tmp_kdata
                first_mol=False
            else:
                if random_overlap and (self.kdatabase.Ng is not None):
                    kdata_array=RandOverlap_2_kdata_prof(self.Narray,
                        self.Nw, self.kdatabase.Ng, 
                        kdata_array,tmp_kdata, self.kdatabase.weights,
                        self.kdatabase.ggrid)
                else:
                    kdata_array+=tmp_kdata

        if rayleigh or (self.cia_database is not None):
            cont_sig=np.zeros((self.Narray,self.Nw))
            if self.cia_database is not None:
                cont_sig+=self.cia_database.cia_cross_section(self.logp_array,
                    self.t_array, vmr_prof, wngrid_limit=self._wn_range)
            if rayleigh:
                if cst_prof:
                    cont_sig+=Rayleigh().sigma(self.wns, self.composition)
                else:
                    cont_sig+=Rayleigh().sigma_array(self.wns, vmr_prof)
            if self.kdatabase.Ng is None:
                kdata_array+=cont_sig
            else:
                kdata_array+=cont_sig[:,:,None]

        return kdata_array
    
    def __getitem__(self, molecule):
        """Overrides getitem
        """
        return self.composition[molecule]

    def __setitem__(self, molecule, vmr):
        """Overrides setitem and recomputes the background gas mixing ratio if needed
        """
        if isinstance(vmr, str):
            self.bg_gas=molecule
            self.composition['inactive_gas']=0.
        else:
            if isinstance(vmr, list):
                self.composition[molecule]=np.array(vmr)
            else:
                self.composition[molecule]=vmr
            if molecule==self.bg_gas:
                self.bg_gas='inactive_gas'
        self.get_background_vmr()
    
    def items(self):
        """Emulates dict.items() method
        """
        return self.composition.items()

    def values(self):
        """Emulates dict.values() method
        """
        return self.composition.values()

    def keys(self):
        """Emulates dict.keys() method
        """
        return self.composition.keys()

    def __repr__(self):
        """Method to output
        """
        output='Volume mixing ratios in Gas_mix: \n'
        for mol, vmr in self.composition.items():
            output+=mol+'->'+str(vmr)+'\n'
        return output

    def copy(self):
        """Deep copy of the dict and arrays. 
        The databases are not deep copied. 
        """
        res=Gas_mix(self.composition, bg_gas=self.bg_gas,
            k_database=self.kdatabase, cia_database=self.cia_database)
        res.logp_array=np.copy(self.logp_array)
        res.t_array=np.copy(self.t_array)
        res.Narray=self.Narray
        res._wn_range=np.copy(self._wn_range)
        res.iw_min=self.iw_min
        res.iw_max=self.iw_max

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers
        """
        if self.wns is not None: return 10000./self.wns

    @property
    def wledges(self):
        """Returns the wavelength array for the bin edges
        """
        if self.wnedges is not None: return 10000./self.wnedges

    def mix_with(self, other_gas, vmr_other_gas):
        """Mix with other Gas_mix.
        """
        raise NotImplementedError()

    def __mul__(self, vmr):    
        """Defines multiplication
        """
        raise NotImplementedError()
        #composition=dict()
        #for mol,mol_vmr in self.composition.items():
        #    composition[mol]=vmr*mol_vmr
        #res=Gas_mix(composition, bg_gas=self.bg_gas)
        #return res

    __rmul__ = __mul__



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
from re import findall
from exo_k.util.singleton import Singleton

class Molar_mass(Singleton):
    """A class to compute the molar mass (in kg/mol) of regular molecules
    with a name written in a regular way (e.g. CO2, H2O, etc.).
    This class can also store the molar mass of custom gases with arbitrary names
    (for example: My_gas, earth_background).
    """
    
    def init(self, *args, **kwds):
        """Initializes empty dictionary of custom molecular masses.
        """
        self._custom_mol_mass={}

    def add_species(self, species_dict):
        """Add one or several species weights to the database. 

        Parameters
        ----------
            species_dict: dict
                Keys are gases names (they do not have to be real molecules).
                Values are molecular weight in kg/mol.
        """
        self._custom_mol_mass.update(species_dict)

    def __repr__(self):
        """Print the currently known species in the database. 
        """
        return self._custom_mol_mass.__repr__()

    def fetch(self, molecule_name):
        """Computes the molar mass of a molecule in kg/mol
        
        Parameters:
            molecule_name: str
                Name of the molecule.

        Returns:
            float:
                Molar mass in kg/mol.
        """
        if molecule_name in self._custom_mol_mass.keys():
            return self._custom_mol_mass[molecule_name]
        #s = re.findall('([A-Z][a-z]?)([0-9]*)', molecule_name)
        s = findall('([A-Za-z][a-z_]*)([0-9]*)', molecule_name)
        molecule_mass = 0
        for element, count in s:
            count = int(count or '1')
            try:
                molecule_mass += _atomic_mass_in_amu[element] * count
            except KeyError:
                print("""The following molecule name was not recognized: {mol}
                    If this is a custom gas
                    add its molecular weight to the library by running:
                    Molar_mass().add_species({{\'{mol}\':weight in kg/mol}})
                    """.format(mol=molecule_name) )
                raise RuntimeError()
        return molecule_mass*1.e-3 # conversion to kg/mol

_atomic_mass_in_amu = {
  "H":	1.00794,
  "He":	4.002602,
  "Li":	6.941,
  "Be":	9.012182,
  "B":	10.811,
  "C":	12.011,
  "N":	14.00674,
  "O":	15.9994,
  "F":	18.9984032,
  "Ne":	20.1797,
  "Na":	22.989768,
  "Mg":	24.3050,
  "Al":	26.981539,
  "Si":	28.0855,
  "P":	30.973762,
  "S":	32.066,
  "Cl":	35.4527,
  "Ar":	39.948,
  "K":	39.0983,
  "Ca":	40.078,
  "Sc":	44.955910,
  "Ti":	47.88,
  "V":	50.9415,
  "Cr":	51.9961,
  "Mn":	54.93805,
  "Fe":	55.847,
  "Co":	58.93320,
  "Ni":	58.6934,
  "Cu":	63.546,
  "Zn":	65.39,
  "Ga":	69.723,
  "Ge":	72.61,
  "As":	74.92159,
  "Se":	78.96,
  "Br":	79.904,
  "Kr":	83.80,
  "Rb":	85.4678,
  "Sr":	87.62,
  "Y":	88.90585,
  "Zr":	91.224,
  "Nb":	92.90638,
  "Mo":	95.94,
  "Tc":	98,
  "Ru":	101.07,
  "Rh":	102.90550,
  "Pd":	106.42,
  "Ag":	107.8682,
  "Cd":	112.411,
  "In":	114.82,
  "Sn":	118.710,
  "Sb":	121.757,
  "Te":	127.60,
  "I":	126.90447,
  "Xe":	131.29,
  "Cs":	132.90543,
  "Ba":	137.327,
  "La":	138.9055,
  "Ce":	140.115,
  "Pr":	140.90765,
  "Nd":	144.24,
  "Pm":	145,
  "Sm":	150.36,
  "Eu":	151.965,
  "Gd":	157.25,
  "Tb":	158.92534,
  "Dy":	162.50,
  "Ho":	164.93032,
  "Er":	167.26,
  "Tm":	168.93421,
  "Yb":	173.04,
  "Lu":	174.967,
  "Hf":	178.49,
  "Ta":	180.9479,
  "W":	183.85,
  "Re":	186.207,
  "Os":	190.2,
  "Ir":	192.22,
  "Pt":	195.08,
  "Au":	196.96654,
  "Hg":	200.59,
  "Tl":	204.3833,
  "Pb":	207.2,
  "Bi":	208.98037,
  "Po":	209,
  "At":	210,
  "Rn":	222,
  "Fr":	223,
  "Ra":	226.0254,
  "Ac":	227,
  "Th":	232.0381,
  "Pa":	213.0359,
  "U":	238.0289,
  "Np":	237.0482,
  "Pu":	244,
  "Am":	243,
  "Cm":	247,
  "Bk":	247,
  "Cf":	251,
  "Es":	252,
  "Fm":	257,
  "Md":	258,
  "No":	259,
  "Lr":	260,
  "Rf":	261,
  "Db":	262,
  "Sg":	263,
  "Bh":	262,
  "Hs":	265,
  "Mt":	266,
}




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A class to handle continuum absorption (CIA)
"""

import os.path
import h5py
import numpy as np
from exo_k.util.filenames import EndOfFile
from .util.interp import linear_interpolation, interp_ind_weights, unit_convert
from .util.cst import KBOLTZ
from .settings import Settings


class Cia_table(object):
    """A class to handle CIA opacity data tables.
    """

    def __init__(self, *filename_filters, filename=None, molecule_pair=None, search_path=None,
            mks=False, remove_zeros=False, old_cia_unit='cm^5'):
        """Initialization for Cia_tables.

        Parameters
        ----------
            filename: str, optional
                Relative or absolute name of the file to be loaded. 
            filename_filters: sequence of string
                As many strings as necessary to uniquely define
                a file in the global search path defined with
                :func:`exo_k.settings.Settings.set_cia_search_path`.
                This path will be searched for a file
                with all the filename_filters in the name.
                The filename_filters can contain '*'.
            molecule_pair: list of size 2, optional
                The molecule pair we want to consider, 
                specified as an array with two strings (like ['H2','H2'] or ['N2','H2O']).
                The order of the molecules in the pair is irrelevant. 

        Other Parameters
        ----------------
            old_cia_unit : str, optional
                String to specify the current cia unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the cia grid and the cia unit do not correspond).
                Available units are: 'cm^5', 'cm^2' that stand for cm^2/amagat,
                and 'cm^-1' that stand for cm^-1/amagat^2.
            remove_zeros: boolean, optional
                If True, the zeros in the kdata table are replaced by
                    a value 10 orders of magnitude smaller than the smallest positive value
            search_path: str, optional
                If search_path is provided,
                it locally overrides the global search path
                defined with :func:`exo_k.settings.Settings.set_cia_search_path`
                and only files in `search_path` are returned.
        """
        self._init_empty()
        self._settings=Settings()
        if filename is not None:
            self.filename=filename
        elif filename_filters or molecule_pair is not None:
            # a none empty sequence returns a True in a conditional statement
            self.filename=self._settings.list_cia_files(*filename_filters,
                molecule_pair=molecule_pair, only_one=True, search_path=search_path)[0]

        if self.filename is not None:
            if self.filename.lower().endswith(('h5','hdf5')):
                self.read_hdf5(self.filename)
            elif self.filename.lower().endswith('cia'):
                self.read_hitran_cia(self.filename, old_cia_unit=old_cia_unit)
            elif self.filename.lower().endswith('.dat'):
                self.read_CKD_cia(self.filename, old_cia_unit=old_cia_unit)
            else:
                raise RuntimeError('Cia file extension not known.')
        if self.abs_coeff is not None:
            if self._settings._convert_to_mks or mks: self.convert_to_mks()
            if remove_zeros : self.remove_zeros()

    def _init_empty(self):
        """Initializes attributes to none.
        """
        self.mol1=None
        self.mol2=None
        self.wns=None
        self.wnedges=None
        self.tgrid=None
        self.abs_coeff=None
        self.abs_coeff_unit='unspecified'
        self.wn_unit='cm^-1'
        self.Nt=None
        self.Nw=None
        self.filename=None


    def read_hitran_cia(self, filename, old_cia_unit='cm^5'):
        """Reads hitran cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
            old_cia_unit: str, optional
                Units found in the file.
        """
        tmp_tgrid=[]
        tmp_abs_coeff=[]
        First=True
        with open(filename, 'r') as file:
            while True:
                try:
                    Nw, Temp = self._read_header(file)
                except EndOfFile:
                    break
                tmp_tgrid.append(Temp)
                tmp_abs_coeff.append([])
                tmp_wns=[]
                if First: 
                    for _ in range(Nw): #as iterator not used, can be replaced by _
                        line=file.readline()
                        tmp=line.split()
                        tmp_wns.append(float(tmp[0]))
                        tmp_abs_coeff[-1].append(float(tmp[1]))
                    self.wns=np.array(tmp_wns)
                else:
                    for _ in range(Nw):
                        line=file.readline()
                        tmp=line.split()
                        tmp_abs_coeff[-1].append(float(tmp[1]))
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.tgrid=np.array(tmp_tgrid)
        self.abs_coeff=np.array(tmp_abs_coeff)
        self.abs_coeff_unit=old_cia_unit
        self.Nt=self.tgrid.size
        self.Nw=self.wns.size


    def _read_header(self, file):
        """Reads the header lines in a Hitran CIA file.

        Parameters
        ----------
            file: file stream
                file to be read.
        """
        line=file.readline()
        if line is None or line=='':
            raise EndOfFile
        tmp = line.split()
        self.mol1,self.mol2 = tmp[0].split('-')
        Nw = int(tmp[3])
        Temp = float(tmp[4])

        return Nw, Temp

    def read_hdf5(self, filename):
        """Reads hdf5 cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
        """
        f = h5py.File(filename, 'r')
        self.wns=f['bin_centers'][...]
        self.abs_coeff=f['abs_coeff'][...]
        self.abs_coeff_unit=f['abs_coeff'].attrs['units']
        self.tgrid=f['t'][...]
        if 'cia_pair' in f:
            self.mol1,self.mol2=f['cia_pair'][()].split('-')
        elif 'cia_pair' in f.attrs:
            self.mol1,self.mol2=f.attrs['cia_pair'].split('-')
        f.close()  
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nt=self.tgrid.size
        self.Nw=self.wns.size
          
    def write_hdf5(self, filename):
        """Writes hdf5 cia files.

        Parameters
        ----------
            filename: str
                Name of the file to be written.
        """
        if not filename.lower().endswith(('.hdf5', '.h5')):
            filename=filename+'.hdf5'
        f = h5py.File(filename, 'w')
        f.create_dataset("bin_centers", data=self.wns,compression="gzip")
        f.create_dataset("t", data=self.tgrid,compression="gzip")
        f.create_dataset("abs_coeff", data=self.abs_coeff,compression="gzip")
        f["abs_coeff"].attrs["units"] = self.abs_coeff_unit
        f.create_dataset("cia_pair", data=self.mol1+'-'+self.mol2)
        f.close()    

    def sample(self, wngrid, remove_zeros=False, use_grid_filter=False, **kwargs):
        """Method to re sample a cia table to a new grid of wavenumbers (in place)

        Parameters
        ----------
            wngrid : array
                new wavenumber grid (cm-1)
            use_grid_filter: boolean, optional
                If true, the table is sampled only within the boundaries
                of its current wavenumber grid. The coefficients are set to zero elswere
                (except if remove_zeros is set to True).
                If false, the values at the boundaries are used when sampling outside the grid.
        """
        wngrid=np.array(wngrid)
        #min_val=np.amin(self.abs_coeff)
        Nnew=wngrid.size
        #wngrid_filter = np.where((wngrid <= self.wnedges[-1]) & (wngrid >= self.wnedges[0]))[0]
        if use_grid_filter:
            wngrid_filter = np.where((wngrid <= self.wns[-1]) & (wngrid >= self.wns[0]))[0]
        else:
            wngrid_filter = np.ones(Nnew,dtype=bool)
        new_abs_coeff=np.zeros((self.Nt,Nnew))
        for iT in range(self.Nt):
            tmp=self.abs_coeff[iT,:]
            new_abs_coeff[iT,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
        self.abs_coeff=new_abs_coeff
        self.wns=wngrid
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nw=Nnew
        if remove_zeros : self.remove_zeros(**kwargs)

    def sample_cp(self, wngrid, **kwargs):
        """Creates a copy of the object before resampling it.

        Parameters
        ----------
            See sample method for details. 

        Returns
        -------
            :class:`Cia_table` object
                the re-sampled :class:`Cia_table`
        """
        res=self.copy()
        res.sample(wngrid, **kwargs)
        return res

    def interpolate_cia(self, t_array=None, log_interp=None, wngrid_limit=None):
        """interpolate_cia interpolates the kdata at on a given temperature profile. 

        Parameters
        ----------
            t_array: float or array
                Temperature array to interpolate to.
                If a float is given, it is interpreted as an array of size 1.
            wngrid_limit: array, optional
                If an array is given, interpolates only within this array.
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).
        """
        if hasattr(t_array, "__len__"):
            t_array=np.array(t_array)
        else:
            t_array=np.array([t_array])
        tind,tweight=interp_ind_weights(t_array,self.tgrid)
        if wngrid_limit is None:
            wngrid_filter = slice(None)
            Nw=self.Nw
        else:
            wngrid_limit=np.array(wngrid_limit)
            wngrid_filter = np.where((self.wnedges > wngrid_limit.min()) & (
                self.wnedges <= wngrid_limit.max()))[0][:-1]
            Nw=wngrid_filter.size
        res=np.zeros((tind.size,Nw))
        if log_interp is None: log_interp=self._settings._log_interp
        if log_interp:
            for ii in range(tind.size):
                kc_t1=np.log(self.abs_coeff[tind[ii]][wngrid_filter].ravel())
                kc_t0=np.log(self.abs_coeff[tind[ii]-1][wngrid_filter].ravel())
                res[ii]=linear_interpolation(kc_t0, kc_t1, tweight[ii])
            return np.exp(res)
        else:
            for ii in range(tind.size):
                #kc_t1=self.abs_coeff[tind[ii]]
                kc_t1=self.abs_coeff[tind[ii]][wngrid_filter].ravel()
                kc_t0=self.abs_coeff[tind[ii]-1][wngrid_filter].ravel()
                res[ii]=linear_interpolation(kc_t0, kc_t1, tweight[ii])
            return res

    def equivalent_xsec(self, logP, T, x_mol2, wngrid_limit=None):
        """Computes the cross section due to CIA in area per molecule of type 1.
        """
        n_density=10**logP/(KBOLTZ*T)
        return self.interpolate_cia(t_array=T,wngrid_limit=wngrid_limit)*n_density*x_mol2

    def effective_cross_section(self, logP, T, x_mol1, x_mol2, wngrid_limit=None):
        """Computes the total cross section for a molecule pair
        (in m^2 per total number of molecules; assumes data in MKS).

        Parameters
        ----------
            logP: float or array
                Log10 of the pressure (Pa).
            T: float or array
                Temperature (K).
            x_mol1/2: float or array
                Volume mixing ratio of the 1st and 2nd molecule of the pair.
            wngrid_limit: array, optional  
                If an array is given, interpolates only within this array. 

        Returns
        -------
            float or array
                total cross section for the molecule pair in m^2 per total number of molecules.
        """
        x_x_n_density=10**logP/(KBOLTZ*T)*x_mol1*x_mol2
        #return self.interpolate_cia( \
        # T,wngrid_limit=wngrid_limit)*n_density[:,None]*n_density[:,None]*x_mol1*x_mol2
        tmp=self.interpolate_cia(t_array=T, wngrid_limit=wngrid_limit)
        return (x_x_n_density*tmp.transpose()).transpose() 
        # trick for the broadcasting to work whether x_x_n_density is a float or an array

    def plot_spectrum(self, ax, t=200., x_axis='wls', xscale=None, yscale=None, **kwarg):
        """Plot the spectrum for a given point

        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            t: float
                temperature(K)
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        toplot=self.interpolate_cia(t)[0]
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        ax.set_ylabel('Abs. coeff')
        ax.grid(True)
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    def convert_abs_coeff_unit(self,abs_coeff_unit='unspecified',old_abs_coeff_unit='unspecified'):
        """Converts abs_coeff to a new unit (in place)

        Parameters
        ----------
            abs_coeff_unit: str
                String to identify the units to convert to.
                Accepts 'cm^5', 'm^5'. or any length^5 unit recognized by the 
                astropy.units library. If ='unspecified', no conversion is done.
            old_abs_coeff_unit : str, optional
                String to specify the current kdata unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the kdata grid and the kdata unit do not correspond)
        """
        if abs_coeff_unit==old_abs_coeff_unit: return
        tmp_k_u_in=old_abs_coeff_unit
        tmp_k_u_out=abs_coeff_unit
        tmp_k_u_file=self.abs_coeff_unit
        self.abs_coeff_unit,conversion_factor=unit_convert( \
            'abs_coeff_unit',unit_file=tmp_k_u_file,unit_in=tmp_k_u_in,unit_out=tmp_k_u_out)
        self.abs_coeff=self.abs_coeff*conversion_factor

    def convert_to_mks(self):
        """Converts units to MKS
        """
        if self.abs_coeff_unit=='cm^5':
            print('Conversion from cm^5 to m^5')
            self.convert_abs_coeff_unit(abs_coeff_unit='m^5')
        elif self.abs_coeff_unit in ['cm^2','m^2']:
            print('Conversion from '+self.abs_coeff_unit+'/amagat to m^5')
            self.convert_abs_coeff_unit(abs_coeff_unit='m^2')
            self.abs_coeff=self.abs_coeff*(KBOLTZ*273.15/101325.0)
            #conversion from m^2/amagat to m^5
        elif self.abs_coeff_unit in ['cm^-1','m^-1']:
            print('Conversion from '+self.abs_coeff_unit+'/amagat^2 to m^5')
            self.convert_abs_coeff_unit(abs_coeff_unit='m^-1')
            self.abs_coeff=self.abs_coeff*(KBOLTZ*273.15/101325.0)**2
            #conversion from m^-1/amagat^2 to m^5
        else:
            pass
        self.abs_coeff_unit='m^5'
        return

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers
        """
        return 10000./self.wns

    def remove_zeros(self, deltalog_min_value=0.):
        """Finds zeros in the abs_coeff and set them to (10.^-deltalog_min_value)
        times the minimum positive value in the table.
        This is to be able to work in logspace. 
        """
        mask = np.zeros(self.abs_coeff.shape,dtype=bool)
        mask[np.nonzero(self.abs_coeff)] = True
        minvalue=np.amin(self.abs_coeff[mask])
        self.abs_coeff[~mask]=minvalue/(10.**deltalog_min_value)
        
    def __repr__(self):
        """Method to output header
        """
        output="""
        file          : {file}
        molecule pair : {mol1} - {mol2}
        t grid   (K)  : {t}
        abs coeff     : {abs_coeff}
        """.format(file=self.filename,mol1=self.mol1, mol2=self.mol2, \
            t=self.tgrid,abs_coeff=self.abs_coeff)
        return output

    def __getitem__(self,key):
        """Overrides getitem.
        """
        return self.abs_coeff[key]

    def copy(self):
        """Creates a new instance of :class:`CIA_table` object and (deep) copies data into it
        """
        res=Cia_table()
        res.mol1 = self.mol1
        res.mol2 = self.mol2
        res.wns = np.copy(self.wns)
        res.wnedges = np.copy(self.wnedges)
        res.tgrid = np.copy(self.tgrid)
        res.abs_coeff = np.copy(self.abs_coeff)
        res.abs_coeff_unit = self.abs_coeff_unit
        res.Nt = self.Nt
        res.Nw = self.Nw
        res.filename = self.filename
        res._settings=Settings()
        return res

    def read_CKD_cia(self, filename, old_cia_unit='cm^2'):
        """Reads hitran cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
        """
        self.tgrid=np.array([200., 250., 300., 350., 400.,
            450., 500., 550., 600., 650., 700.])
        self.Nt=self.tgrid.size
        self.abs_coeff=np.loadtxt(filename, skiprows=1, unpack=True)
        self.mol1='H2O'
        if 'SELF' in os.path.basename(filename):
            self.mol2='H2O'
        else:
            self.mol2='N2'
        nu_name=os.path.join(os.path.dirname(filename),'H2O_CONT_NU.dat')
        self.wns=np.loadtxt(nu_name, skiprows=1, unpack=True)
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.abs_coeff_unit=old_cia_unit
        self.Nw=self.wns.size

#    #  amagatS=(273.15/temp)*(presS/101325.0)
#    #  amagatF=(273.15/temp)*(presF/101325.0)
#
#    #  abcoef = abcoefS*amagatS + abcoefF*amagatF ! Eq. (15) in Clough (1989)
#    #  abcoef = abcoef*(presS/(presF+presS))      ! take H2O mixing ratio into account
#    #                                                ! abs coeffs are given per molecule of H2O
#
#    #  Nmolec = (presS+presF)/(kB*temp)           ! assume ideal gas
#!      print*,'Total number of molecules per m^3 is',Nmolec
#
#      abcoef = (abcoefS*x_H2O + abcoefF*(1-x_other))*(k_b*273.15/101325.0)*x_H2O
#                     *Nmolec*2/(100.0**2)          ! convert to m^-1

    def effective_cross_section2(self, logP, T, x_mol1, x_mol2, wngrid_limit=None):
        """Obsolete.
        
        Computes the total cross section for a molecule pair
        (in m^2 per total number of molecules; assumes data in MKS).
        """
        x_x_n_density=10**logP/(KBOLTZ*T)*x_mol1*x_mol2
        #return self.interpolate_cia( \
        # T,wngrid_limit=wngrid_limit)*n_density[:,None]*n_density[:,None]*x_mol1*x_mol2
        tmp=self.interpolate_cia(t_array=T, wngrid_limit=wngrid_limit)
        return x_x_n_density[:,None]*tmp




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
from math import log10
import os
import h5py
import numpy as np
import astropy.units as u
from scipy.interpolate import RegularGridInterpolator
from .data_table import Data_table
from .util.interp import rm_molec, rebin_ind_weights, \
        gauss_legendre, spectrum_to_kdist, is_sorted, bin_down_corrk_numba
from .util.cst import KBOLTZ
from .hires_spectrum import Hires_spectrum
from .util.filenames import create_fname_grid_Kspectrum_LMDZ, select_kwargs


class Ktable5d(Data_table):
    """A class that handles tables of k-coefficients with a variable gas.
    Based on the Data_table class that handles basic operations common to Ktable and Xtable.

    This class is specifically designed to deal with
    LMDZ type ktable where there is a variable gas.
    """
    
    def __init__(self, *filename_filters, filename=None, path=None,
        p_unit='unspecified', file_p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        remove_zeros=False, search_path=None, mol=None,
        **kwargs):
        """Initializes k coeff table with variable gas and
        supporting data from various sources (see below by order of precedence)

        Parameters
        ----------
            filename: str (optional)
                Relative or absolute name of the file to be loaded. 
            path: str
                If none of the above is specifed,
                path can point to a directory with a LMDZ type k coeff table.
                In this case, see read_LMDZ for the keywords to specify.

        If there is no input, just creates an empty object to be filled later

        See :class:`~exo_k.ktable.Ktable` __init__ mehthod for documentation on
        `p_unit`, `file_p_unit`, `kdata_unit`, `file_kdata_unit`, `remove_zeros`,
        `search_path`, and `mol` arguments.

        """
        super().__init__()

        if filename is not None:
            self.filename=filename
        elif filename_filters or (mol is not None and path is None): 
        # a none empty sequence returns a True in a conditional statement
            self.filename=self._settings.list_files(*filename_filters, molecule = mol,
                only_one=True, search_path=search_path)[0]
        if self.filename is not None:
            if self.filename.lower().endswith(('.hdf5', '.h5')):
                self.read_hdf5(filename=self.filename, mol=mol)
            else:
                raise NotImplementedError( \
                    'Requested format not recognized. Should end with .pickle, .hdf5, or .h5')
        elif path is not None:
            self.read_LMDZ(path=path, mol=mol, **kwargs)
        else:                  #if there is no input file, just create an empty object 
            self.wnedges=None
            self.weights=None
            self.ggrid=None
            self.gedges=None
            self.xgrid=None
            self._finterp_kdata=None

        super().finalize_init(p_unit=p_unit, file_p_unit=file_p_unit,
            kdata_unit=kdata_unit, file_kdata_unit=file_kdata_unit,
            remove_zeros=remove_zeros)

        if self.kdata is not None:
            self.setup_interpolation()

    @property
    def shape(self):
        """Returns the shape of self.kdata
        """
        return np.array([self.Np,self.Nt,self.Nx,self.Nw,self.Ng])

    def read_hdf5(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from an Exomol hdf5 file

        Parameters
        ----------
            filename : str
                Name of the input hdf5 file
        """
        if (filename is None or not filename.lower().endswith(('.hdf5', '.h5'))):
            raise RuntimeError("You should provide an input hdf5 file")
        f = h5py.File(filename, 'r')
        if 'mol_name' in f:
            self.mol=f['mol_name'][()]
        elif 'mol_name' in f.attrs:
            self.mol=f.attrs['mol_name']
        else:
            self.mol=os.path.basename(filename).split(self._settings._delimiter)[0]
        if isinstance(self.mol, np.ndarray): self.mol=self.mol[0]
        if mol is not None: self.mol=mol
        self.wns=f['bin_centers'][...]
        self.wnedges=f['bin_edges'][...]
        if 'units' in f['bin_edges'].attrs:
            self.wn_unit=f['bin_edges'].attrs['units']
        else:
            if 'units' in f['bin_centers'].attrs:
                self.wn_unit=f['bin_centers'].attrs['units']
        self.kdata=f['kcoeff'][...]
        self.kdata_unit=f['kcoeff'].attrs['units']
        self.tgrid=f['t'][...]
        self.pgrid=f['p'][...]
        self.logpgrid=np.log10(self.pgrid)
        self.p_unit=f['p'].attrs['units']
        self.xgrid=f['x'][...]
        if 'weights' in f.keys():
            self.weights=f['weights'][...]
        else:
            raise RuntimeError('No weights keyword. This file is probably a cross section file.')
        self.ggrid=f['samples'][...]
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.logk=False
        f.close()  
        self.Np,self.Nt,self.Nx,self.Nw,self.Ng=self.kdata.shape

    def write_hdf5(self, filename):
        """Saves data in a hdf5 format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.hdf5', '.h5')):
            fullfilename=filename+'.hdf5'
        compression="gzip"
        f = h5py.File(fullfilename, 'w')
        f.create_dataset("mol_name", data=self.mol)
        f.create_dataset("p", data=self.pgrid,compression=compression)
        f["p"].attrs["units"] = self.p_unit
        f.create_dataset("t", data=self.tgrid,compression=compression)
        f.create_dataset("x", data=self.xgrid,compression=compression)
        f.create_dataset("kcoeff", data=self.kdata,compression=compression)
        f["kcoeff"].attrs["units"] = self.kdata_unit
        f.create_dataset("samples", data=self.ggrid,compression=compression)
        f.create_dataset("weights", data=self.weights,compression=compression)
        f.create_dataset("bin_edges", data=self.wnedges,compression=compression)
        f.create_dataset("bin_centers", data=self.wns,compression=compression)
        f["bin_edges"].attrs["units"] = self.wn_unit
        f.close()    

    def read_LMDZ(self, path=None, res=None, band=None, mol=None):
        """Initializes k coeff table and supporting data from a .dat file
        in a gcm friendly format.

        Units are assumed to be cm^2 for kdata and mbar for pressure. 

        Parameters
        ----------
            path: str
                Name of the directory with the various input files
            res: str
                "IRxVI" where IR and VI are the numbers of bands
                in the infrared and visible of the k table to load.
            band: str
                "IR" or "VI" to specify which band to load.
        """        
        if (path is None) or (res is None): \
            raise TypeError("You should provide an input directory name and a resolution")

        self.filename=path

        self.weights=np.loadtxt(os.path.join(path,'g.dat'),skiprows=1)[:-1]
        # we remove the last point that is always zero.
        # in the gcm this last point is intended to take care of continuum
        self.Ng=self.weights.size
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5

        self.p_unit='mbar'
        self.logpgrid=np.loadtxt(os.path.join(path,'p.dat'),skiprows=1)*1.
        self.Np=self.logpgrid.size
        self.pgrid=10**self.logpgrid

        self.tgrid=np.loadtxt(os.path.join(path,'T.dat'),skiprows=1)
        self.Nt=self.tgrid.size

        _, self.mol, self.Nx, self.xgrid = read_Qdat(os.path.join(path,'Q.dat'))
        if mol is not None: self.mol=mol

        if band is None:
            raw=np.loadtxt(os.path.join(path,res,'narrowbands.in'), skiprows=1, unpack=True)
        else:
            raw=np.loadtxt(os.path.join(path,res,'narrowbands_'+band+'.in'), \
                skiprows=1, unpack=True)
        self.wnedges=np.append(raw[0],raw[1,-1])
        self.wns=(self.wnedges[1:]+self.wnedges[:-1])*0.5
        self.Nw=self.wns.size
        
        self.kdata_unit='cm^2/molecule'
        if band is None:
            file_to_load=os.path.join(path,res,'corrk_gcm.dat')
        else:
            file_to_load=os.path.join(path,res,'corrk_gcm_'+band+'.dat')        
        tmp=np.loadtxt(file_to_load) \
            .reshape((self.Nt,self.Np,self.Nx,self.Nw,self.Ng+1),order='F')
        self.kdata=tmp[:,:,:,:,:-1].transpose((1,0,2,3,4))  
        # also removing the last g point which is equal to 0.
        self.logk=False        
        return None

    def write_LMDZ(self, path, band='IR', fmt='%22.15e', write_only_metadata=False):
        """Saves data in a LMDZ friendly format.
        
        The gcm requires p in mbar and kdata in cm^2/molec.
        The conversion is done automatically.

        Parameters
        ----------
            path: str
                Name of the directory to be created and saved,
                the one that will contain all the necessary files
            band: str
                The band you are computing: 'IR' or 'VI'
            fmt: str
                Fortran format for the corrk file. 
            write_only_metadata: bool, optional
                If `True`, only supporting files are written (T.dat, p.dat, etc.)
        """
        try:
            os.mkdir(path)
        except FileExistsError:
            print('Directory was already there: '+path)
        file = open(os.path.join(path,'p.dat'), "w")
        file.write(str(self.Np)+'\n')
        lp_to_write=self.logpgrid+np.log10(u.Unit(self.p_unit).to(u.Unit('mbar')))
        for lp in lp_to_write:
            file.write(str(lp)+'\n')
        file.close()

        file = open(os.path.join(path,'T.dat'), "w")
        file.write(str(self.Nt)+'\n')
        for t in self.tgrid:
            file.write(str(t)+'\n')
        file.close()

        file = open(os.path.join(path,'g.dat'), "w")
        file.write(str(self.Ng+1)+'\n')
        for g in self.weights:
            file.write(str(g)+'\n')
        file.write(str(0.)+'\n')
        file.close()

        dirname=os.path.join(path,band+str(self.Nw))
        try:
            os.mkdir(dirname)
        except FileExistsError:
            print('Directory was already there: '+dirname)

        file = open(os.path.join(dirname,'narrowbands_'+band+'.in'), "w")
        file.write(str(self.Nw)+'\n')
        for iw in range(self.Nw):
            file.write(str(self.wnedges[iw])+' '+str(self.wnedges[iw+1])+'\n')
        file.close()

        if not write_only_metadata:
            #file = open(dirname+'/corrk_gcm_IR.in', "w")
            data_to_write=self.kdata.transpose((1,0,2,3,4)).flatten(order='F')
            data_to_write=data_to_write*u.Unit(rm_molec(self.kdata_unit)).to(u.Unit('cm^2'))
            data_to_write=np.append(data_to_write, \
                np.zeros(self.Np*self.Nt*self.Nx*self.Nw)) \
                    .reshape((1,self.Np*self.Nt*self.Nx*self.Nw*(self.Ng+1)))
            np.savetxt(os.path.join(dirname,'corrk_gcm_'+band+'.dat'),data_to_write,fmt=fmt)

    def hires_to_ktable(self, path=None, filename_grid=None,
        logpgrid=None, tgrid=None, xgrid=None, wnedges=None,
        quad='legendre', order=20, weights=None, ggrid=None,
        mid_dw=True, write=0, mol=None,
        grid_p_unit='Pa', p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        **kwargs):
        """Computes a k coeff table from :class:`~exo_k.util.hires_spectrum.Hires_spectrum`
        objects.

        see :func:`exo_k.ktable.Ktable.hires_to_ktable` method for details
        on the arguments and options.

        Other Parameters
        ----------------
            xgrid: array
                Input grid in vmr of the variable gas. Needed for a Ktable5d.

        .. warning::
            By default, log pressures are specified in Pa in logpgrid!!! If you want
            to use another unit, do not forget to specify it with the grid_p_unit keyword.

        """        
        if path is None: raise TypeError("You should provide an input hires_spectrum directory")
        if wnedges is None: raise TypeError("You should provide an input wavenumber array")

        self.filename=path
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        if weights is not None:
            self.weights=weights
            self.gedges=np.concatenate(([0],np.cumsum(self.weights)))
            if ggrid is not None: 
                self.ggrid=np.array(ggrid)
            else:
                self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5
        else:
            if quad=='legendre':
                self.weights,self.ggrid,self.gedges=gauss_legendre(order)
            else:
                raise NotImplementedError("Type of quadrature (quad keyword) not known.")
        self.Ng=self.weights.size

        conversion_factor=u.Unit(grid_p_unit).to(u.Unit('Pa'))
        self.logpgrid=np.array(logpgrid, dtype=float)+np.log10(conversion_factor)
        self.pgrid=10**self.logpgrid #in Pa
        self.p_unit='Pa'
        self.Np=self.logpgrid.size
        if write >= 3 : print(self.Np,self.pgrid)

        self.tgrid=np.array(tgrid)
        self.Nt=self.tgrid.size
        if write >= 3 : print(self.Nt,self.tgrid)

        self.xgrid=np.array(xgrid)
        self.Nx=self.xgrid.size
        if write >= 3 : print(self.Nx,self.xgrid)

        self.wnedges=np.array(wnedges)
        if self.wnedges.size<2: raise TypeError('wnedges should at least have two values')
        self.wns=(self.wnedges[1:]+self.wnedges[:-1])*0.5
        self.Nw=self.wns.size
        
        self.kdata=np.zeros(self.shape)
        if filename_grid is None:
            filename_grid=create_fname_grid_Kspectrum_LMDZ(self.Np, self.Nt,
                **select_kwargs(kwargs,['suffix','nb_digit']))
        else:
            filename_grid=np.array(filename_grid)

        for iP in range(self.Np):
          for iT in range(self.Nt):
            for iX in range(self.Nx):
                filename=filename_grid[iP,iT,iX]
                fname=os.path.join(path,filename)
                if write >= 3 : print(fname)

                spec_hr=Hires_spectrum(fname, file_kdata_unit=file_kdata_unit,
                    **select_kwargs(kwargs,['skiprows','wn_column','mult_factor',
                        'kdata_column','data_type']))
                # for later conversion, the real kdata_unit is in spec_hr.kdata_unit
                self.kdata_unit=spec_hr.kdata_unit
                was_xsec=(spec_hr.data_type=='xsec')
                wn_hr=spec_hr.wns
                k_hr=spec_hr.kdata
                if mid_dw:
                    dwn_hr=(wn_hr[2:]-wn_hr[:-2])*0.5
                    wn_hr=wn_hr[1:-1]
                    k_hr=k_hr[1:-1]
                else:
                    dwn_hr=(wn_hr[1:]-wn_hr[:-1])
                    wn_hr=wn_hr[:-1]
                    k_hr=k_hr[:-1]
                self.kdata[iP,iT,iX]=spectrum_to_kdist(k_hr,wn_hr,dwn_hr,self.wnedges,self.ggrid)
                if not was_xsec:
                    self.kdata[iP,iT,iX]=self.kdata[iP,iT,iX]*KBOLTZ*self.tgrid[iT]/self.pgrid[iP]
        if not was_xsec:
            self.kdata=self.kdata*u.Unit(self.kdata_unit).to(u.Unit('m^-1'))
            self.kdata_unit='m^2/molecule'
            # Accounts for the conversion of the abs_coeff to m^-1, so we know that
            #  self.kdata is now in m^2/molecule. Can now convert to the desired unit.
        if self._settings._convert_to_mks and kdata_unit is 'unspecified':
            kdata_unit='m^2/molecule'
        self.convert_kdata_unit(kdata_unit=kdata_unit)
        # converts from self.kdata_unit wich is either:
        #   - 'm^2/molecule' data_type was 'abs_coeff'
        #   - The units after Hires_spectrum (which have already taken
        #        file_kdata_unit into account)
        #  to the desired unit.
        self.convert_p_unit(p_unit=p_unit)

    def setup_interpolation(self, log_interp=None):
        """Creates interpolating functions to be called later on.
        """
        if log_interp is None: log_interp=self._settings._log_interp
        self._local_log_interp=log_interp
        if self._local_log_interp:
            self._finterp_kdata=RegularGridInterpolator( \
                (self.logpgrid,self.tgrid,np.log(self.xgrid)), np.log(self.kdata), \
                bounds_error=True)
        else:
            self._finterp_kdata=RegularGridInterpolator( \
                (self.logpgrid,self.tgrid,np.log(self.xgrid)), self.kdata, \
                bounds_error=True)

    def set_kdata(self, new_kdata):
        """Changes kdata. this is preferred to directly accessing kdata because one
        could forget to run setup_interpolation().

        Parameters
        ----------
            new_kdata: array
                New array of kdata.
        """
        self.kdata=new_kdata
        self.setup_interpolation()

    def interpolate_kdata(self, logp_array=None, t_array=None, x_array= None,
            log_interp=None, wngrid_limit=None):
        """interpolate_kdata interpolates the kdata at on a given temperature and
        log pressure profile. 

        Parameters
        ----------
            logp_array: Array
                log 10 pressure array to interpolate to
            t_array: Array, same size a logp_array
                Temperature array to interpolate to
            x_array: Array, same size a logp_array
                vmr of variable gas array to interpolate to
            If floats are given, they are interpreted as arrays of size 1.
            wngrid_limit: list or array, optional
                if an array is given, interpolates only within this array
            log_interp: bool, dummy
                Dummy variable to be consistent with interpolate_kdata in data_table.
                Whether the interpolation is linear in kdata or in log(kdata) is actually
                controlled by self._settings._log_interp but only when the ktable is loaded.
                If you change that after the loading, you should rerun setup_interpolation().

        """
        coord_to_interp=np.array([logp_array,t_array,np.log(x_array)]).transpose()
        tmp_res=self._finterp_kdata(coord_to_interp)
        if wngrid_limit is None:
            wngrid_filter = slice(None)
        else:
            wngrid_limit=np.array(wngrid_limit)
            wngrid_filter = np.where((self.wnedges > wngrid_limit.min()) & (
                self.wnedges <= wngrid_limit.max()))[0][:-1]
        if self._local_log_interp:
            return np.exp(tmp_res[:,wngrid_filter])
        else:
            return tmp_res[:,wngrid_filter]


    def remap_logPT(self, logp_array=None, t_array=None, x_array= None):
        """remap_logPT re-interpolates the kdata on a new temprature and log pressure grid. 

        Parameters
        ----------
            logp_array: Array
                log 10 pressure array to interpolate to
            t_array: Array
                temperature array to interpolate to
            x_array: Array
                vmr of variable gas array to interpolate to

        Whether the interpolation is linear in kdata or in log(kdata)
        is controlled by self._settings._log_interp but only when the ktable is loaded.
        If you change that after the loading, you should rerun setup_interpolation().
        """
        coord=np.array(np.meshgrid(logp_array, t_array, np.log(x_array))).transpose((2,1,3,0))
        if self._local_log_interp:
            tmp_res=np.exp(self._finterp_kdata(coord))
        else:
            tmp_res=self._finterp_kdata(coord)
        self.logpgrid= logp_array
        self.pgrid   = 10**self.logpgrid
        self.tgrid   = t_array
        self.xgrid   = x_array
        self.Np      = logp_array.size
        self.Nt      = t_array.size
        self.Nx      = x_array.size
        self.set_kdata(tmp_res)
        self.setup_interpolation()

    def copy(self,cp_kdata=True):
        """Creates a new instance of :class:`Ktable5d` object and (deep) copies data into it

        Parameters
        ----------
            cp_kdata: bool, optional
                If false, the kdata table is not copied and
                only the structure and metadata are. 

        Returns
        -------
            :class:`Ktable`
                A new :class:`Ktable5d` instance with the same structure as self.
        """
        res=Ktable5d()
        res.copy_attr(self,cp_kdata=cp_kdata)
        res._local_log_interp=self._local_log_interp
        res.xgrid   = np.copy(self.xgrid)
        res.weights = np.copy(self.weights)
        res.ggrid   = np.copy(self.ggrid)
        res.gedges  = np.copy(self.gedges)
        if res.kdata is not None: res.setup_interpolation()
        return res

    def gindex(self, g):
        """Finds the index corresponding to the given g
        """
        return min(np.searchsorted(self.ggrid,g),self.Ng-1)

    def xindex(self, x):
        """Finds the index corresponding to the given x
        """
        return min(np.searchsorted(self.xgrid,x),self.Nx-1)

    def spectrum_to_plot(self, p=1.e-5, t=200., x=1., g=None):
        """provide the spectrum for a given point to be plotted

        Parameters
        ----------
            p : float
                Pressure (Ktable pressure unit)
            t : float
                Temperature(K)
            g: float
                Gauss point
            x: float
                Mixing ratio of the species
        """
        if g is None: raise RuntimeError('A gauss point should be provided with the g= keyword.')
        gindex=self.gindex(g)
        return self.interpolate_kdata(log10(p),t,x)[0,:,gindex]

    def plot_distrib(self, ax, p=1.e-5, t=200., wl=1., x=1., xscale=None, yscale='log', **kwarg):
        """Plot the distribution for a given point

        Parameters
        ----------
            p : float
                Pressure (Ktable pressure unit)
            t : float
                Temperature(K)
            wl: float
                Wavelength (micron)
        """
        wlindex=self.wlindex(wl)
        toplot=self.interpolate_kdata(log10(p),t,x)[0,wlindex]
        if xscale is not None: 
            ax.set_xscale(xscale)
            ax.plot(1.-self.ggrid,toplot,**kwarg)
            ax.set_xlabel('1-g')
        else:
            ax.plot(self.ggrid,toplot,**kwarg)
            ax.set_xlabel('g')
        ax.set_ylabel('Cross section ('+self.kdata_unit+')')
        ax.grid(True)
        if yscale is not None: ax.set_yscale(yscale)

    def __repr__(self):
        """Method to output header
        """
        out1=super().__repr__()
        output=out1+"""
        weights      : {wg}
        x      (vmr) : {xgrid}
        data oredered following p, t, x, wn, g
        shape        : {shape}
        """.format(wg=self.weights, xgrid=self.xgrid, shape=self.shape)
        return output

    def combine_with(self, other, x_self=None, x_other=None, **kwargs):
        """Method to create a new :class:`Ktable5d` where the kdata of 'self' are
        randomly mixed with 'other' (that must be a :class:`Ktable`).

        The main purpose is to add the opacity of a trace species to the background gas
        of the :class:`Ktable5d` instance. 

        .. warning::
            Because:
            
              * the opacity from the background and variable gases cannot be
                isolated,
              * The values of the array for the vmr of the variable gas (self.xgrid)
                are not modified (diluted),

            the treatment here is valid only if `x_other` << 1.
            
            For this reason, `x_self` should be either left to None, or 1-`x_other` depending
            exactly what you want to do. But if you put `x_self`<<1, you are on your own.

        Parameters
        ----------
            other: :class:`~exo_k.ktable.Ktable`
                A :class:`Ktable` object to be mixed with. Dimensions should be the same as self
                (except for xgrid).
            x_self: float only, optional
                Volume mixing ratio of self.
            x_other: float or array, optional
                Volume mixing ratio of the species to be mixed with (other).

        If either x_self or x_other are set to `None` (default),
        the cross section of the species in question
        are considered to be already normalized with respect to the mixing ratio.

        Returns
        -------
            :class:`Ktable5d`
                A new Ktable5d with the opacity of the new species added. 
        """
        if not ((self.Np == other.Np) and (self.Nt == other.Nt) and (self.Nw == other.Nw) \
            and (self.Ng == other.Ng)):
            raise TypeError("""in combine_with: kdata tables do not have the same dimensions.
                I'll stop now!""")
        if other.Nx is not None:
            raise TypeError("""in combine_with: cannot combine 2 Ktable5d""")
        if (self.p_unit!=other.p_unit) or \
            (rm_molec(self.kdata_unit)!=rm_molec(other.kdata_unit)):
            raise RuntimeError("""in combine_with: tables do not have the same units.
                I'll stop now!""")
        res=self.copy(cp_kdata=True)
        tmp=other.copy()
        if x_other is None:
            x_other=1.
        for iX in range(self.Nx):
            tmp.kdata=self.kdata[:,:,iX,:,:]
            res.kdata[:,:,iX,:,:]= \
                tmp.RandOverlap(other, x_self, x_other*(1.-self.xgrid[iX]), **kwargs)
        res.setup_interpolation()
        return res

    def bin_down(self, wnedges=None, weights=None, ggrid=None,
        remove_zeros=False, num=300, use_rebin=False, write=0):
        """Method to bin down a kcoeff table to a new grid of wavenumbers

        Parameters
        ----------
            wnedges: array
                Edges of the new bins of wavenumbers (cm-1)
                onto which the kcoeff should be binned down.
                if you want Nwnew bin in the end, wnedges.size must be Nwnew+1
                wnedges[0] should be greater than self.wnedges[0] (JL20 not sure anymore)
                wnedges[-1] should be lower than self.wnedges[-1]
            weights: array, optional
                Desired weights for the resulting Ktable.
            ggrid: array, optional
                Desired g-points for the resulting Ktable.
                Must be consistent with provided weights.
                If not given, they are taken at the midpoints of the array
                given by the cumulative sum of the weights
        """
        old_ggrid=self.ggrid
        if weights is not None:
            self.weights=np.array(weights)
            self.gedges=np.concatenate(([0],np.cumsum(self.weights)))
            if ggrid is not None: 
                self.ggrid=np.array(ggrid)
            else:
                self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5
            self.Ng=self.ggrid.size
        wnedges=np.array(wnedges)
        if wnedges.size<2: raise TypeError('wnedges should at least have two values')
        wngrid_filter = np.where((wnedges <= self.wnedges[-1]) & (wnedges >= self.wnedges[0]))[0]
        if not is_sorted(wnedges):
            raise RuntimeError('wnedges should be sorted.')
        indicestosum,wn_weigths=rebin_ind_weights(self.wnedges, wnedges[wngrid_filter])
        if write> 10 :print(indicestosum);print(wn_weigths)

        newkdata=np.zeros((self.Np,self.Nt,self.Nx,wnedges.size-1,self.Ng), dtype=np.float64)
        for iX in range(self.Nx):
            newkdata[:,:,iX]=bin_down_corrk_numba((self.Np,self.Nt,wnedges.size-1,self.Ng), \
                self.kdata[:,:,iX], old_ggrid, self.ggrid, self.gedges, indicestosum, \
                wngrid_filter, wn_weigths, num, use_rebin)
        self.kdata=newkdata
        self.wnedges=wnedges
        self.wns=(wnedges[1:]+wnedges[:-1])*0.5
        self.Nw=self.wns.size
        if remove_zeros : self.remove_zeros(deltalog_min_value=10.)
        self.setup_interpolation()
    
    def clip_spectral_range(self, wn_range=None, wl_range=None):
        """Limits the data to the provided spectral range:

           * Wavenumber in cm^-1 if using wn_range argument
           * Wavelength in micron if using wl_range
        """
        super().clip_spectral_range(wn_range=wn_range, wl_range=wl_range)
        if self.kdata is not None:
            self.setup_interpolation()


def read_Qdat(filename):
    """Reads Q.dat files LMDZ style and extract the vmr grid.

    Parameters
    ----------
        filename: str
            Path to file to read.
    Returns
    -------
        background_mol_names: list
            list of names of molecules in background gas
        var_mol: str
            Name of variable molecule
        Nx: int
            Size of xgrid
        xgrid: array
            grid of vmr for variable gas
    """
    file = open(filename, "r")
    Nmol=int(file.readline())
    background_mol_names=[]
    for ii in range(Nmol-1):
        #print(file.readline().split()[0])
        background_mol_names.append(file.readline().split()[0])
    var_mol=file.readline().split()[0]
    Nx=int(file.readline())
    xgrid=np.zeros(Nx)
    for ii in range(Nx):
        xgrid[ii]=float(file.readline())
    return background_mol_names,var_mol,Nx,xgrid



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Module containing a class to handle a database of CIA tables to compute opacities with it.
"""
import numpy as np

from .cia_table import Cia_table
from .settings import NoFileFoundError

class CIAdatabase(object):
    """Class to group :class:`~exo_k.cia_table.Cia_table` objects
    and combine them in radiative transfer
    """

    def __init__(self, *str_filters, filenames=None, molecule_pairs=None,
        molecules=None, remove_zeros=True, **kwargs):
        """Initializes cia tables and supporting data.
        The files to load can be specified either using (in order of precedence)
        a list of filenames, molecule pairs, or molecules. See below for details.

        Any number of filters (strings) can be provided to refine the search to
        files whose names contain the filters. 

        Parameters
        ----------
            filenames: list, optional
                List of names (not full path) of the input cia files. 
                The files must be in the global search path.
            molecule_pairs: list of size 2 lists, optional
                List of the molecule pairs we want to consider, 
                specified as an array with two strings (like ['H2','H2'] or ['N2','H2O']).
                The order of the molecules in the pair is irrelevant. 
            molecules: list of strings, optional
                A list of all the molecules we want to consider. `Exo_k` will look
                for all the possible pairs for which a cia file is found. 

        The default path searched is set with :func:`exo_k.settings.Settings.set_cia_search_path`
        or :func:`~exo_k.settings.Settings.add_cia_search_path`.
        A local search path can be specified with the `search_path` keyword.

        See the options of :class:`~exo_k.cia_table.Cia_table` __init__ method.
        """
        self.cia_tables={}
        self.wns=None
        self.Nw=None
        self.abs_coeff_unit=None
        if filenames is not None:
            for filename in filenames:
                tmp_cia_table=Cia_table(*([filename]+list(str_filters)),
                    remove_zeros=remove_zeros, **kwargs)
                self.add_cia_tables(tmp_cia_table)
        elif molecule_pairs is not None:
            for mol_pair in molecule_pairs:
                tmp_cia_table=Cia_table(*str_filters, molecule_pair=mol_pair,
                    remove_zeros=remove_zeros, **kwargs)
                self.add_cia_tables(tmp_cia_table)
        elif molecules is not None:
            for mol1 in molecules:
                for mol2 in molecules:
                    try:
                        tmp_cia_table=Cia_table(*str_filters, molecule_pair=[mol1,mol2],
                            remove_zeros=remove_zeros, **kwargs)
                        self.add_cia_tables(tmp_cia_table)
                    except NoFileFoundError:
                        pass


    
    def add_cia_tables(self, *cia_tables):
        """Adds new :class:`~exo_k.cia_table.Cia_table` objects to a CIA database.

        Parameters
        ----------
            cia_tables: :class:`~exo_k.cia_table.Cia_table`
                As many cia tables as you want.
        """
        for cia_table in cia_tables:
            if self.abs_coeff_unit is None:
                self.abs_coeff_unit=cia_table.abs_coeff_unit
            elif self.abs_coeff_unit!=cia_table.abs_coeff_unit:
                print('CIAdatabase units for cia coeff: {k}'.format(k=self.abs_coeff_unit))
                print('{mol1}-{mol2} units for cia coeff: {k}'.format(\
                    mol1=cia_table.mol1, mol2=cia_table.mol2, k=cia_table.abs_coeff_unit))
                raise RuntimeError("""You naughty:
                all Cia_tables in a database must have the same coeff units""")

            if cia_table.mol1 in self.cia_tables:
                if cia_table.mol2 in self.cia_tables[cia_table.mol1]:
                    continue
                else:
                    self.cia_tables[cia_table.mol1][cia_table.mol2]=cia_table
            else:
                self.cia_tables[cia_table.mol1]={cia_table.mol2:cia_table}

    def __getitem__(self, molecule):
        """Overrides getitem so that CIAdatabase['mol'] directly accesses 
        the database for that molecule.
        """
        if molecule not in self.cia_tables:
            raise KeyError('The requested molecule is not available.')
        return self.cia_tables[molecule]

    def copy(self):
        """Creates a new instance of :class:`CIAdatabase` object and (deep) copies data into it
        """
        res=CIAdatabase(None)
        for dic_cia_tabs in self.cia_tables.values():
            for cia_tab in dic_cia_tabs.values():
                res.add_cia_tables(cia_tab.copy())
        return res

    def __repr__(self):
        """Method to output
        """
        output='The available molecule pairs are: \n'
        for mol1, dico in self.cia_tables.items():
            for mol2, cia_tab in dico.items():
                output+=mol1+'-'+mol2+'->'+cia_tab.filename+'\n'
        if self.wns is not None:
            output+='All tables share a common spectral grid\n'
        else:
            output+='All tables do NOT have common spectral grid\n'
            output+='You will need to run sample before using the database\n'

        return output

    def sample(self, wngrid, remove_zeros=False, use_grid_filter=False):
        """Samples all the cia_table in the database on the same wavenumber grid
        to be able to use them in radiative transfer modules.

        Parameters
        ----------
            wngrid : array
                new wavenumber grid (cm-1)

        .. seealso::
            See :func:`exo_k.cia_table.Cia_table.sample for further details on options.

        """
        for mol1 in self.cia_tables.values():
            for cia_table in mol1.values():
                cia_table.sample(wngrid, remove_zeros=remove_zeros,
                    use_grid_filter=use_grid_filter)
        self.wns=np.array(wngrid)
        self.Nw=self.wns.size

    def convert_to_mks(self):
        """Converts units of all Cia_tables to MKS.
        """
        first=True
        for mol1 in self.cia_tables.values():
            for cia_table in mol1.values():
                cia_table.convert_to_mks()
                if first:
                    self.abs_coeff_unit=cia_table.abs_coeff_unit
        

    def cia_cross_section(self, logP_array, T_array, gas_comp, wngrid_limit=None):
        """Computes the absorption coefficient in m^-1 for the whole mix specified 
        (assumes data in MKS).

        Parameters
        ----------
            logP_array: array

            T_array: array
                log10 Pressure and temperature profiles

            gas_comp: :class:`~exo_k.gas_mix.Gas_mix` object
                behaves like a dict with mol names as keys and vmr as values.

            wngrid_limit: array, optional
                Smaller and bigger wavenumbers inside which to perform the calculation.

        Returns:
            array:
                The cia effective cross section coefficient profile for the whole gas (in m^2).
        """
        logP_array=np.array(logP_array)
        T_array=np.array(T_array)
        first=True
        if self.wns is None: raise RuntimeError("""cia tables must be sampled
            on the same grid before calling cia_cross_section.
            Please use the sample() method.""")
        for mol,x1 in gas_comp.items():
            if mol in self.cia_tables.keys():
                for mol2 in self.cia_tables[mol].keys():
                    if mol2 in gas_comp.keys():
                        x2=gas_comp[mol2]
                        if first:
                            res=self.cia_tables[mol][mol2].effective_cross_section( \
                                logP_array, T_array, x1, x2, wngrid_limit=wngrid_limit)
                            first=False
                        else:
                            res+=self.cia_tables[mol][mol2].effective_cross_section( \
                                logP_array, T_array, x1, x2, wngrid_limit=wngrid_limit)
        if first: # means that no molecule was in the database, we need to initialize res
            res=np.zeros((T_array.size,self.wns.size))
        return res

        




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
__init__ module to load the exo_k library
"""
from exo_k.util.user_func import *
from exo_k.util.radiation import *
from exo_k.util.cst import *
from exo_k.util.spectrum import Spectrum
from exo_k.util.filenames import *
from exo_k.util.molar_mass import Molar_mass
from .ktable import Ktable
from .ktable5d import Ktable5d
from .kdatabase import Kdatabase
from .xtable import Xtable
from .atable import Atable, combine_tables
from .adatabase import Adatabase
from .cia_table import Cia_table
from .ciadatabase import CIAdatabase
from .hires_spectrum import Hires_spectrum
from .atm import Atm_profile,Atm
from .chemistry import EquChemTable
from .gas_mix import Gas_mix
from .settings import Settings
from .rayleigh import Rayleigh



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Library of useful functions for users (only).

Functions here CANNOT be called into the library itself: importing this module in others would
lead to recursive import problems.
"""
import os
import numpy as np
from exo_k.ktable import Ktable
from exo_k.ktable5d import Ktable5d
from exo_k.xtable import Xtable
from exo_k.hires_spectrum import Hires_spectrum
from exo_k.util.cst import PI

def hires_to_ktable(filename_grid=None, xgrid=None, **kwargs):
    """Emulates :func:`exo_k.ktable.Ktable.hires_to_ktable`
    and :func:`exo_k.ktable5d.Ktable5d.hires_to_ktable` as functions and not methods
    so that the user can call them without creating a Ktable first.

    See those methods for details on the available arguments and options.
    """
    if filename_grid is not None:
        if len(np.array(filename_grid).shape) == 3 and xgrid is None:
            raise RuntimeError("""From the shape of filename_grid
                it seems you want have an xgrid dimension, but xgrid is None!
                """)    
    if xgrid is not None: # creates a Ktable5d
        if filename_grid is not None:
            if len(np.array(filename_grid).shape) != 3:
                raise RuntimeError("""You provided a xgrid of vmrs, 
                    but the shape of filename_grid is not 
                    of the type (Np, Nt, Nx)!
                    """)
        res=Ktable5d()
        res.hires_to_ktable(filename_grid=filename_grid, xgrid=xgrid, **kwargs)
    else:
        res=Ktable()
        res.hires_to_ktable(filename_grid=filename_grid, **kwargs)
    return res

def hires_to_xtable(**kwargs):
    """Emulates :func:`exo_k.xtable.Xtable.hires_to_xtable`
    as a function and not a method
    so that the user can call it without creating an Xtable first.

    See :func:`~exo_k.xtable.Xtable.hires_to_xtable` for details
    on the available arguments and options.
    """
    res=Xtable()
    res.hires_to_xtable(**kwargs)
    return res

def convert_kspectrum_to_hdf5(file_in, file_out=None, **kwargs):
    """Converts kspectrum like spectra to hdf5 format for speed and space.
    Helper function. Real work done in :class:`~exo_k.util.hires_spectrum.Hires_spectrum`
    __init__ funtion.

    Go there to see all the available arguments and options.

    Parameters
    ----------
        file_in: str
            Initial kspectrum filename.
        file_out: str
            Name of the final hdf5 file to be created. If not provided,
            'file_in.h5' will be used. 
    """
    tmp=Hires_spectrum(file_in, **kwargs)
    if file_out is None: file_out=file_in
    tmp.write_hdf5(file_out)

def create_fname_grid(base_string, logpgrid=None, tgrid=None, xgrid=None,
        p_kw=None, t_kw=None, x_kw=None):
    """Creates a grid of filenames from an array of pressures, temperatures (
    and vmr if there is a variable gas).

    Parameters
    ----------
        base_string: str
            Generic name of the spectra files with specific keywords to be replaced 
            by the relevant numerical values
        logpgrid: Array
            Grid in log(pressure/Pa) of the input
        tgrid: Array
            Grid in temperature of the input
        xgrid: Array
            Input grid in vmr of the variable gas
        
    .. warning::
        The result of this function is much more predictable 
        if the values in the above arrays are given as integers. 
        If you want to use floats anyway, good luck. 

    Parameters
    ----------
        p_kw: str
        t_kw: str
        x_kw: str
            The pattern string that will be recognized as keywords between
            {} in base_string (See examples).

    Examples
    --------

        >>> logpgrid=[1,2]
        >>> tgrid=np.array([100.,150.,200.])
        >>> file_grid=exo_k.create_fname_grid('spectrum_CO2_1e{logp}Pa_{t}K.hdf5',
                  logpgrid=logpgrid,tgrid=tgrid,p_kw='logp',t_kw='t')                  
        array([['spectrum_CO2_1e1Pa_100K.hdf5', 'spectrum_CO2_1e1Pa_150K.hdf5',
        'spectrum_CO2_1e1Pa_200K.hdf5'],
        ['spectrum_CO2_1e2Pa_100K.hdf5', 'spectrum_CO2_1e2Pa_150K.hdf5',
        'spectrum_CO2_1e2Pa_200K.hdf5']], dtype='<U28')

    """
    logpgrid=np.array(logpgrid)
    tgrid=np.array(tgrid)
    res=[]
    if xgrid is None:
        for iP in range(logpgrid.size):
            for iT in range(tgrid.size):
                dict_opt={p_kw:str(logpgrid[iP]),t_kw:str(tgrid[iT])}
                fname=base_string.format(**dict_opt)
                res.append(fname)
        return np.array(res).reshape((logpgrid.size,tgrid.size))
    else:
        xgrid=np.array(xgrid)
        for iP in range(logpgrid.size):
            for iT in range(tgrid.size):
                for iX in range(xgrid.size):
                    dict_opt={p_kw:str(logpgrid[iP]), \
                        t_kw:str(tgrid[iT]),x_kw:str(xgrid[iX])}
                    fname=base_string.format(**dict_opt)
                    res.append(fname)
        return np.array(res).reshape((logpgrid.size,tgrid.size,xgrid.size))

def finalize_LMDZ_dir(corrkname, IRsize, VIsize):
    """Creates the right links for a LMDZ type directory to be read by the LMDZ generic GCM.

    You will need to create a proper Q.dat before using with the LMDZ GCM. 

    .. important::
        You must have run :func:`exo_k.ktable.Ktable.write_LMDZ` or
        :func:`exo_k.ktable5d.Ktable5d.write_LMDZ` for both of your IR and VI channels
        beforehand.

    Parameters
    ----------
        corrkname: str
            Path to the directory with the LMDZ ktable to finalize
        IRsize: int
            Number of IR spectral bins
        VIsize: int
            Number of VI spectral bins
    """
    newdir=os.path.join(corrkname,str(IRsize)+'x'+str(VIsize))
    try:
        os.mkdir(newdir)
    except FileExistsError:
        os.system('rm -rf '+newdir)
        os.mkdir(newdir)
    os.symlink('../IR'+str(IRsize)+'/corrk_gcm_IR.dat',os.path.join(newdir,'corrk_gcm_IR.dat'))
    os.symlink('../IR'+str(IRsize)+'/narrowbands_IR.in',os.path.join(newdir,'narrowbands_IR.in'))
    os.symlink('../VI'+str(VIsize)+'/corrk_gcm_VI.dat',os.path.join(newdir,'corrk_gcm_VI.dat'))
    os.symlink('../VI'+str(VIsize)+'/narrowbands_VI.in',os.path.join(newdir,'narrowbands_VI.in'))
    print('Everything went ok. Your ktable is in:',newdir)
    print("You'll probably need to add Q.dat before using it though!")

def mmr_to_number_density(mmr, gas_density, r_eff, condensate_density):
    """Converts a mass mixing ratio (mmr or q) in a number density of particles
    (in number per unit volume)

    Parameters
    ----------
        mmr: float or array
            Mass mixing ratio (in kg per kg of air)
        gas_density: float or array
            Density of the gas (in kg/m^3)
        r_eff: float or array
            Effective radius of the particles
        condensate_density: float or array
            Density of the constituent of the condensed particles (in kg/m^3)
    """
    particle_mass=4.*PI*r_eff**3*condensate_density/3.
    return mmr*gas_density/particle_mass
    